#if 1
/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Benchmarks for authenticated ciphers.
 *
 * @version $Id$
 * @ingroup bench
 */

#ifdef __ICC430__
#include <msp430.h>
#endif

#include <stdio.h>

#include "relic.h"
#include "relic_bench.h"

static const unsigned int SIZES[] = {10 * 1024};

static void enc_ecb(bc_ctx_t ctx, size_t n) {
	bc_t msg;
	bc_t cipher;

	for (int i = 0; i < n / BC_BLOCK_LEN; i++) {
		bc_enc(ctx, cipher, msg);
	}
	if (n % BC_BLOCK_LEN) {
		bc_enc(ctx, cipher, msg);
	}
}

static void dec_ecb(bc_ctx_t ctx, size_t n) {
	bc_t msg;
	bc_t cipher;

	for (int i = 0; i < n / BC_BLOCK_LEN; i++) {
		bc_dec(ctx, cipher, msg);
	}
	if (n % BC_BLOCK_LEN) {
		bc_dec(ctx, cipher, msg);
	}
}

static void enc_ctr(bc_ctx_t ctx, size_t n) {
	align unsigned char msg[SIZES[0]];
	align unsigned char cipher[SIZES[0]];

	int i;
	for (i = 0; i < n; i += BC_BLOCK_LEN) {
		bc_ctr_enc(ctx, cipher + i, msg + i, BC_BLOCK_LEN);
	}
	if (n % BC_BLOCK_LEN) {
		bc_ctr_enc(ctx, cipher + i, msg + i, n % BC_BLOCK_LEN);
	}
}

static void dec_ctr(bc_ctx_t ctx, size_t n) {
	align unsigned char msg[SIZES[0]];
	align unsigned char cipher[SIZES[0]];

	int i;
	for (i = 0; i < n; i += BC_BLOCK_LEN) {
		bc_ctr_dec(ctx, cipher + i, msg + i, BC_BLOCK_LEN);
	}
	if (n % BC_BLOCK_LEN) {
		bc_ctr_dec(ctx, cipher + i, msg + i, n % BC_BLOCK_LEN);
	}
}

static void enc(ac_ctx_t ctx, unsigned char *key, unsigned int key_len, unsigned char iv[AC_IV_LEN],
		size_t msg_len, size_t data_len) {
	align unsigned char msg[SIZES[0]];
	align unsigned char cipher[SIZES[0]];
	unsigned char tag[AC_TAG_LEN];

	ac_init(ctx, key, BC_KEY_LEN, iv, AC_IV_LEN, msg_len, 0);

	int i;
	for (i = 0; i < msg_len; i += AC_BLOCK_LEN) {
		ac_enc(ctx, cipher + i, msg + i, sizeof(msg));
	}
	if (msg_len % AC_BLOCK_LEN) {
		ac_enc(ctx, cipher + i, msg + i, msg_len % AC_BLOCK_LEN);
	}
	ac_tag(ctx, tag, sizeof(tag));
}

static void dec(ac_ctx_t ctx, unsigned char *key, unsigned int key_len, unsigned char iv[AC_IV_LEN],
		size_t msg_len, size_t data_len) {
	unsigned char msg[SIZES[0]];
	unsigned char cipher[SIZES[0]];
	unsigned char tag[AC_TAG_LEN];

	ac_init(ctx, key, BC_KEY_LEN, iv, AC_IV_LEN, msg_len, 0);

	int i;
	for (i = 0; i < msg_len; i += AC_BLOCK_LEN) {
		ac_dec(ctx, msg + i, cipher + i, sizeof(msg));
	}
	if (msg_len % AC_BLOCK_LEN) {
		ac_dec(ctx, msg + i, cipher + i, msg_len % AC_BLOCK_LEN);
	}
	ac_check(ctx, tag, sizeof(tag));
}

static void authenticated_encryption(void) {
	unsigned char key[BC_KEY_LEN];
	unsigned char iv[AC_IV_LEN];
	ac_ctx_t ctx;

	for (int k = 0; k < sizeof(SIZES) / sizeof(int); k++) {
		ac_key(ctx, key, sizeof(key));
		//forces OCB3 to cache key init
		ac_init(ctx, key, sizeof(key), iv, sizeof(iv), SIZES[k], 0);
		BENCH_BEGIN("authenticated encryption") {
			BENCH_ADD(enc(ctx, key, sizeof(key), iv, SIZES[k], 0));
		}
		BENCH_END;

		ac_key(ctx, key, sizeof(key));
		ac_init(ctx, key, sizeof(key), iv, sizeof(iv), SIZES[k], 0);
		BENCH_BEGIN("decryption-verification") {
			BENCH_ADD(dec(ctx, key, sizeof(key), iv, SIZES[k], 0));
		}
		BENCH_END;
	}
}

static void encryption(void) {
	unsigned char key[BC_KEY_LEN];
	bc_ctx_t ctx;

	for (int k = 0; k < sizeof(SIZES) / sizeof(int); k++) {
		bc_enc_key(ctx, key, sizeof(key));
		BENCH_BEGIN("encryption") {
			BENCH_ADD(enc_ctr(ctx, SIZES[k]));
		}
		BENCH_END;

		bc_enc_key(ctx, key, sizeof(key));
		BENCH_BEGIN("decryption") {
			BENCH_ADD(dec_ctr(ctx, SIZES[k]));
		}
		BENCH_END;
	}
}

#undef BENCH
#define BENCH 2048

static void gcm_mul(void) {
	align dig_t a[128/sizeof(dig_t)];
	align dig_t b[128/sizeof(dig_t)];
	align dig_t c[128/sizeof(dig_t)];
	BENCH_BEGIN("gcm_mul") {
		BENCH_ADD(ac_gcm_mul_low(c, a, b));
	}
	BENCH_END;
}

int main(void) {
#ifdef __ICC430__
	// Stop watchdog timer to prevent time out reset
	WDTCTL = WDTPW + WDTHOLD;

	UCSCTL3 |= SELREF_2;// Set DCO FLL reference = REFO
	UCSCTL4 |= SELA_2;// Set ACLK = REFO

	__bis_SR_register(SCG0);// Disable the FLL control loop
	UCSCTL0 = 0x0000;// Set lowest possible DCOx, MODx
	UCSCTL1 = DCORSEL_7;// Select DCO range 24MHz operation
	UCSCTL2 = FLLD_1 + 609;// Set DCO Multiplier for 20 MHz
						   // (N + 1) * FLLRef = Fdco
						   // (609 + 1) * 32768 = 20 MHz
						   // Set FLL Div = fDCOCLK/2
	__bic_SR_register(SCG0);// Enable the FLL control loop

	// Worst-case settling time for the DCO when the DCO range bits have been
	// changed is n x 32 x 32 x f_MCLK / f_FLL_reference. See UCS chapter in 5xx
	// UG for optimization.
	// 32 x 32 x 20 MHz / 32,768 Hz = 375000 = MCLK cycles for DCO to settle
	__delay_cycles(625000);
#endif
	core_init();
	conf_print();
	util_banner("Benchmarks for the AC module:", 0);

	gcm_mul();

	encryption();

	authenticated_encryption();

	core_clean();
	return 0;
}
#endif
