/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Interface of the low-level authenticated cipher module.
 *
 * @version $Id$
 * @ingroup ac
 */

#ifndef RELIC_AC_LOW_H
#define RELIC_AC_LOW_H

/*
 * Provides useful constants for Assembly backends.
 */
#ifdef ASM

#include "relic_conf.h"

#else

#include "relic_types.h"

/**
 * Computes the xor of two blocks for CCFBH mode.
 *
 * Each block must have AC_CCFBH_BLOCK_LEN bytes.
 *
 * @param[out] c			- the result.
 * @param[in] a				- the first block to xor.
 * @param[in] b				- the second block to xor.
 */
void ac_ccfbh_xor_low(unsigned char *c, unsigned char *a, unsigned char *b);

/**
 * Computes the xor of two tag-sized blocks for CCFBH mode.
 *
 * Each block must have AC_CCFBH_TAG_LEN bytes.
 *
 * @param[out] c			- the first block and result.
 * @param[in] a				- the second block to xor.
 */
void ac_ccfbh_xortag_low(unsigned char *c, unsigned char *a);

/**
 * Copies a block for CCFBH mode.
 *
 * Each block must have AC_CCFBH_BLOCK_LEN bytes.
 *
 * @param[out] c			- the destination block.
 * @param[in] a				- the block to copy.
 */
void ac_ccfbh_copy_low(unsigned char *c, unsigned char *a);

/**
 * Builds a precomputation table for binary field multiplication.
 *
 * @param[out] t			- the table.
 * @param[in] h				- the base element.
 */
void ac_gcm_tab_low(dig_t *t, unsigned char *h);

/**
 * Multiplies a block by H using the precomputed table.
 * @param[out] c			- the result.
 * @param[in] a				- the block.
 * @param[in] b				- the table.
 */
void ac_gcm_mul_low(dig_t *c, dig_t *a, dig_t *b);

/**
 * The Hummingbird-2 WD16 function.
 *
 * @param[in] xa			- the `x` value xored to `a`.
 * @param b					- the `b` value.
 * @param c					- the `c` value.
 * @param d					- the `d` value.
 * @return WD16(x, a, b, c, d).
 */
uint16_t ac_hb2_wd16_low(uint16_t xa, uint16_t b, uint16_t c, uint16_t d);

/**
 * The inverse of the Hummingbird-2 WD16 function.
 *
 * The result must be xored with `a` to compute the original x.
 *
 * @param[in] y				- an output of the WD16 function.
 * @param b					- the `b` value used in WD16.
 * @param c					- the `c` value used in WD16.
 * @param d					- the `d` value used in WD16.
 * @return the original `x` xored with `a`.
 */
uint16_t ac_hb2_wd16inv_low(uint16_t y, uint16_t b, uint16_t c, uint16_t d);

/**
 * Computes the OCB3 "double" function.
 *
 * Multiplies by x in the OCB3 finite field.
 *
 * @param b[in,out]			- the input block, also receives the result.
 */
void ac_ocb3_dbl_low(bc_t b);

/**
 * Computes the number of trailing (least significant) zeros in the binary
 * representation of the input.
 *
 * @param v[in]				- the input.
 * @return the number of trailing zeros in the input.
 */
unsigned int ac_ocb3_ntz_low(unsigned int v);

#endif /* !ASM */

#endif /* !RELIC_AC_LOW_H */
