/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @defgroup ac Authenticated ciphers.
 */

/**
 * @file
 *
 * Interface of the module for authenticated encryption ciphers.
 *
 * @version $Id$
 * @ingroup bn
 */

#ifndef RELIC_AC_H
#define RELIC_AC_H

#include <stddef.h>

#include "relic_bc.h"
#include "relic_ac_conf.h"
#include "relic_types.h"

/*============================================================================*/
/* Constant definitions                                                       */
/*============================================================================*/

#define AC_CCFBH_KEY_LEN BC_KEY_LEN
#define AC_CCM_KEY_LEN BC_KEY_LEN
#define AC_GCM_KEY_LEN BC_KEY_LEN
#define AC_HB2_KEY_LEN 16
#define AC_MASHA_KEY_LEN 16
#define AC_OCB3_KEY_LEN BC_KEY_LEN

/**
 * Key length in bytes.
 */
#if AC_METHD == CCFBH
#define AC_KEY_LEN AC_CCFBH_KEY_LEN
#elif AC_METHD == CCM
#define AC_KEY_LEN AC_CCM_KEY_LEN
#elif AC_METHD == GCM
#define AC_KEY_LEN AC_GCM_KEY_LEN
#elif AC_METHD == HB2
#define AC_KEY_LEN AC_HB2_KEY_LEN
#elif AC_METHD == MASHA
#define AC_KEY_LEN AC_MASHA_KEY_LEN
#elif AC_METHD == OCB3
#define AC_KEY_LEN AC_OCB3_KEY_LEN
#endif

//Some tag lengths are defined in the configuration.
#define AC_GCM_TAG_LEN 16
#define AC_HB2_TAG_LEN 16
#define AC_MASHA_TAG_LEN 32
#define AC_OCB3_TAG_LEN 16

/**
 * Tag length in bytes.
 */
#if AC_METHD == CCFBH
#define AC_TAG_LEN AC_CCFBH_TAG_LEN
#elif AC_METHD == CCM
#define AC_TAG_LEN AC_CCM_TAG_LEN
#elif AC_METHD == GCM
#define AC_TAG_LEN AC_GCM_TAG_LEN
#elif AC_METHD == HB2
#define AC_TAG_LEN AC_HB2_TAG_LEN
#elif AC_METHD == MASHA
#define AC_TAG_LEN AC_MASHA_TAG_LEN
#elif AC_METHD == OCB3
#define AC_TAG_LEN AC_OCB3_TAG_LEN
#endif

#define AC_CCFBH_BLOCK_LEN (BC_BLOCK_LEN - AC_CCFBH_TAG_LEN)
#define AC_CCM_BLOCK_LEN BC_BLOCK_LEN
#define AC_GCM_BLOCK_LEN BC_BLOCK_LEN
#define AC_HB2_BLOCK_LEN 2
#define AC_MASHA_BLOCK_LEN 8
#define AC_OCB3_BLOCK_LEN BC_BLOCK_LEN

/**
 * Block length in bytes.
 */
#if AC_METHD == CCFBH
#define AC_BLOCK_LEN AC_CCFBH_BLOCK_LEN
#elif AC_METHD == CCM
#define AC_BLOCK_LEN AC_CCM_BLOCK_LEN
#elif AC_METHD == GCM
#define AC_BLOCK_LEN AC_GCM_BLOCK_LEN
#elif AC_METHD == HB2
#define AC_BLOCK_LEN AC_HB2_BLOCK_LEN
#elif AC_METHD == MASHA
#define AC_BLOCK_LEN AC_MASHA_BLOCK_LEN
#elif AC_METHD == OCB3
#define AC_BLOCK_LEN AC_OCB3_BLOCK_LEN
#endif

#define AC_CCFBH_IV_LEN AC_BLOCK_LEN
#define AC_CCM_IV_LEN 12
#define AC_GCM_IV_LEN 12
#define AC_HB2_IV_LEN 12
#define AC_MASHA_IV_LEN 24
#define AC_OCB3_IV_LEN 12

/**
 * Initialization vector length in bytes.
 */
#if AC_METHD == CCFBH
#define AC_IV_LEN AC_CCFBH_IV_LEN
#elif AC_METHD == CCM
#define AC_IV_LEN AC_CCM_IV_LEN
#elif AC_METHD == GCM
#define AC_IV_LEN AC_GCM_IV_LEN
#elif AC_METHD == HB2
#define AC_IV_LEN AC_HB2_IV_LEN
#elif AC_METHD == MASHA
#define AC_IV_LEN AC_MASHA_IV_LEN
#elif AC_METHD == OCB3
#define AC_IV_LEN AC_OCB3_IV_LEN
#endif

/*============================================================================*/
/* Type definitions                                                           */
/*============================================================================*/

/**
 * Represents the context of the CCFB+H encryption / decryption process.
 */
typedef struct {
	bc_t last_c;
	align unsigned char tag[AC_TAG_LEN];
	unsigned int *i;
	bc_ctx_t bc_ctx;
} ac_ccfbh_ctx_st;

/**
 * Pointer to a CCFB+H context.
 */
typedef ac_ccfbh_ctx_st ac_ccfbh_ctx_t[1];

/**
 * Represents the context of the CCM encryption / decryption process.
 */
typedef struct {
	bc_t last_y;
	bc_t s0;
	bc_t data_buf;
	size_t data_buf_len;
	bc_ctx_t bc_ctx;
} ac_ccm_ctx_st;

/**
 * Pointer to a CCM context.
 */
typedef ac_ccm_ctx_st ac_ccm_ctx_t[1];

/**
 * Represents the context of the GCM encryption / decryption process.
 */
typedef struct {
	bc_t last_y;
	dig_t len_a;
	dig_t len_c;
	dig_t table[AC_GCM_TABLE * (AC_GCM_BLOCK_LEN / sizeof(dig_t) + 1)];
	bc_ctx_t bc_ctx;
} ac_gcm_ctx_st;

/**
 * Pointer to a CCM context.
 */
typedef ac_gcm_ctx_st ac_gcm_ctx_t[1];

/**
 * Represents the context of the Hummingbird-2 encryption / decryption process.
 */
typedef struct {
	uint16_t K[8];
	uint16_t R[8];
	uint16_t IV[4];
} ac_hb2_ctx_st;

/**
 * Pointer to a Hummingbird-2 context.
 */
typedef ac_hb2_ctx_st ac_hb2_ctx_t[1];

/**
 * Represents the context of the MASHA encryption / decryption process.
 */
typedef struct {
	uint32_t A[16];
	uint32_t B[43];
	uint32_t R1;
	uint32_t R2;
	uint32_t L1;
	uint32_t L2;
	int32_t clock;
	uint32_t ciphertext[2];
	int complete;
} ac_masha_ctx_st;

/**
 * Pointer to a MASHA context.
 */
typedef ac_masha_ctx_st ac_masha_ctx_t[1];

/**
 * Represents the context of the OCB3 encryption / decryption process.
 */
typedef struct {
	bc_t offset;
	bc_t checksum;
	bc_t Lstar;
	bc_t Ldollar;
	bc_t Ltmp;
	bc_t L[8];
	bc_t ad_checksum;
	bc_t ad_offset;
	bc_t cached_Top;
	dig_t KtopStr[(2 * BC_BLOCK_LEN) / sizeof(dig_t)];
	unsigned ad_blocks_processed;
	unsigned blocks_processed;
#ifdef BC_SINGL
	unsigned char key[BC_KEY_LEN];
	int config; //0=enc, 1=dec
#endif
	bc_ctx_t enc_ctx;
	bc_ctx_t dec_ctx;
} ac_ocb3_ctx_st;

/**
 * Pointer to a OCB3 context.
 */
typedef ac_ocb3_ctx_st ac_ocb3_ctx_t[1];

/**
 * Represents the context of authenticated encryption / decryption process.
 */
#if AC_METHD == CCM
typedef ac_ccm_ctx_st ac_ctx_st;
#elif AC_METHD == GCM
typedef ac_gcm_ctx_st ac_ctx_st;
#elif AC_METHD == OCB3
typedef ac_ocb3_ctx_st ac_ctx_st;
#elif AC_METHD == MASHA
typedef ac_masha_ctx_st ac_ctx_st;
#elif AC_METHD == HB2
typedef ac_hb2_ctx_st ac_ctx_st;
#elif AC_METHD == CCFBH
typedef ac_ccfbh_ctx_st ac_ctx_st;
#endif

/**
 * Pointer to the context of authenticated encryption / decryption process.
 */
typedef ac_ctx_st ac_ctx_t[1];

/**
 * Sets the authenticated encryption / decryption key.
 *
 * @param[out] ctx			- the context.
 * @param[in] key			- the key.
 * @param[in] key_len		- the length of the key in bytes.
 */
#if AC_METHD == CCFBH
#define ac_key(ctx, key, key_len) ac_ccfbh_key(ctx, key, key_len)
#elif AC_METHD == CCM
#define ac_key(ctx, key, key_len) ac_ccm_key(ctx, key, key_len)
#elif AC_METHD == GCM
#define ac_key(ctx, key, key_len) ac_gcm_key(ctx, key, key_len)
#elif AC_METHD == HB2
#define ac_key(ctx, key, key_len) ac_hb2_key(ctx, key, key_len)
#elif AC_METHD == MASHA
#define ac_key(ctx, key, key_len) ac_masha_key(ctx, key, key_len)
#elif AC_METHD == OCB3
#define ac_key(ctx, key, key_len) ac_ocb3_key(ctx, key, key_len)
#endif

void ac_ccfbh_key(ac_ccfbh_ctx_t ctx, unsigned char *key, size_t key_len);

void ac_ccm_key(ac_ccm_ctx_t ctx, unsigned char *key, size_t key_len);

void ac_gcm_key(ac_gcm_ctx_t ctx, unsigned char *key, size_t key_len);

void ac_hb2_key(ac_hb2_ctx_t ctx, unsigned char *key, size_t key_len);

void ac_masha_key(ac_masha_ctx_t ctx, unsigned char *key, size_t key_len);

void ac_ocb3_key(ac_ocb3_ctx_t ctx, unsigned char *key, size_t key_len);

/**
 * Initializes authenticated encryption / decryption with an initialization
 * vector, preparing to send a message.
 *
 * Some modes do not require the key; in that case, it can be NULL.
 * However, in order to write generic code, the same key set before must be
 * used here.
 *
 * The msg_len and data_len must be specified for modes which are not online
 * (e.g. CCM). Otherwise, they can be zero.
 *
 * @param[out] ctx			- the context.
 * @param[in] key			- the key
 * @param[in] key_len		- the key length in bytes.
 * @param[in] iv			- the initialization vector (IV).
 * @param[in] iv_len		- the IV length in bytes.
 * @param[in] msg_len		- the length in bytes of the message to be processed.
 * @param[in] data_len		- the length in bytes of additional data to be processed.
 */
#if AC_METHD == CCFBH
#define ac_init(ctx, key, key_len, iv, iv_len, msg_len, data_len) \
	ac_ccfbh_init(ctx, key, key_len, iv, iv_len, msg_len, data_len)
#elif AC_METHD == CCM
#define ac_init(ctx, key, key_len, iv, iv_len, msg_len, data_len) \
	ac_ccm_init(ctx, key, key_len, iv, iv_len, msg_len, data_len)
#elif AC_METHD == GCM
#define ac_init(ctx, key, key_len, iv, iv_len, msg_len, data_len) \
	ac_gcm_init(ctx, key, key_len, iv, iv_len, msg_len, data_len)
#elif AC_METHD == HB2
#define ac_init(ctx, key, key_len, iv, iv_len, msg_len, data_len) \
	ac_hb2_init(ctx, key, key_len, iv, iv_len, msg_len, data_len)
#elif AC_METHD == MASHA
#define ac_init(ctx, key, key_len, iv, iv_len, msg_len, data_len) \
	ac_masha_init(ctx, key, key_len, iv, iv_len, msg_len, data_len)
#elif AC_METHD == OCB3
#define ac_init(ctx, key, key_len, iv, iv_len, msg_len, data_len) \
	ac_ocb3_init(ctx, key, key_len, iv, iv_len, msg_len, data_len)
#endif

void ac_ccfbh_init(ac_ccfbh_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len);

void ac_ccm_init(ac_ccm_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len);

void ac_gcm_init(ac_gcm_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len);

void ac_hb2_init(ac_hb2_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len);

void ac_masha_init(ac_masha_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len);

void ac_ocb3_init(ac_ocb3_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len);

/**
 * Inputs additional data to be authenticated only.
 *
 * The data buffer must be aligned. It must have BC_BLOCK_LEN bytes, except for
 * the last call when it can have 0 < data_len <= BC_BLOCK_LEN.
 *
 * @param[in,out] ctx		- the context.
 * @param[in] data			- the additional data.
 * @param[in] data_len		- the length in bytes of the data.
 */
#if AC_METHD == CCFBH
#define ac_data(ctx, data, data_len) ac_ccfbh_data(ctx, data, data_len)
#elif AC_METHD == CCM
#define ac_data(ctx, data, data_len) ac_ccm_data(ctx, data, data_len)
#elif AC_METHD == GCM
#define ac_data(ctx, data, data_len) ac_gcm_data(ctx, data, data_len)
#elif AC_METHD == HB2
#define ac_data(ctx, data, data_len) ac_hb2_data(ctx, data, data_len)
#elif AC_METHD == MASHA
#define ac_data(ctx, data, data_len) ac_masha_data(ctx, data, data_len)
#elif AC_METHD == OCB3
#define ac_data(ctx, data, data_len) ac_ocb3_data(ctx, data, data_len)
#endif

void ac_ccfbh_data(ac_ccfbh_ctx_t ctx, unsigned char *data, size_t data_len);

void ac_ccm_data(ac_ccm_ctx_t ctx, unsigned char *data, size_t data_len);

void ac_gcm_data(ac_gcm_ctx_t ctx, unsigned char *data, size_t data_len);

void ac_hb2_data(ac_hb2_ctx_t ctx, unsigned char *data, size_t data_len);

void ac_masha_data(ac_masha_ctx_t ctx, unsigned char *data, size_t data_len);

void ac_ocb3_data(ac_ocb3_ctx_t ctx, unsigned char *data, size_t data_len);

/**
 * Encrypts plaintext.
 *
 * The input buffer must be aligned. It must have BC_BLOCK_LEN bytes, except for
 * the last call when it can have 0 < input_len <= BC_BLOCK_LEN.
 *
 * @param[in,out] ctx		- the context.
 * @param[out] output		- the ciphertext generated.
 * @param[in] input			- hte plaintext to encrypt.
 * @param[in] input_len		- the length in bytes of the plaintext.
 */
#if AC_METHD == CCFBH
#define ac_enc(ctx, output, input, input_len) ac_ccfbh_enc(ctx, output, input, input_len)
#elif AC_METHD == CCM
#define ac_enc(ctx, output, input, input_len) ac_ccm_enc(ctx, output, input, input_len)
#elif AC_METHD == GCM
#define ac_enc(ctx, output, input, input_len) ac_gcm_enc(ctx, output, input, input_len)
#elif AC_METHD == HB2
#define ac_enc(ctx, output, input, input_len) ac_hb2_enc(ctx, output, input, input_len)
#elif AC_METHD == MASHA
#define ac_enc(ctx, output, input, input_len) ac_masha_enc(ctx, output, input, input_len)
#elif AC_METHD == OCB3
#define ac_enc(ctx, output, input, input_len) ac_ocb3_enc(ctx, output, input, input_len)
#endif

void ac_ccfbh_enc(ac_ccfbh_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

void ac_ccm_enc(ac_ccm_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

void ac_gcm_enc(ac_gcm_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

void ac_hb2_enc(ac_hb2_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

void ac_masha_enc(ac_masha_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

void ac_ocb3_enc(ac_ocb3_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

/**
 * Decrypts ciphertext.
 *
 * The input buffer must be aligned. It must have BC_BLOCK_LEN bytes, except for
 * the last call when it can have 0 < input_len <= BC_BLOCK_LEN.
 *
 * @param[in,out] ctx		- the context.
 * @param[out] output		- the plaintext generated.
 * @param[in] input			- hte ciphertext to decrypt.
 * @param[in] input_len		- the length in bytes of the ciphertext.
 */
#if AC_METHD == CCFBH
#define ac_dec(ctx, output, input, input_len) ac_ccfbh_dec(ctx, output, input, input_len)
#elif AC_METHD == CCM
#define ac_dec(ctx, output, input, input_len) ac_ccm_dec(ctx, output, input, input_len)
#elif AC_METHD == GCM
#define ac_dec(ctx, output, input, input_len) ac_gcm_dec(ctx, output, input, input_len)
#elif AC_METHD == HB2
#define ac_dec(ctx, output, input, input_len) ac_hb2_dec(ctx, output, input, input_len)
#elif AC_METHD == MASHA
#define ac_dec(ctx, output, input, input_len) ac_masha_dec(ctx, output, input, input_len)
#elif AC_METHD == OCB3
#define ac_dec(ctx, output, input, input_len) ac_ocb3_dec(ctx, output, input, input_len)
#endif

void ac_ccfbh_dec(ac_ccfbh_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

void ac_ccm_dec(ac_ccm_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

void ac_gcm_dec(ac_gcm_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

void ac_hb2_dec(ac_hb2_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

void ac_masha_dec(ac_masha_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

void ac_ocb3_dec(ac_ocb3_ctx_t ctx, unsigned char *output, unsigned char *input, size_t input_len);

/**
 * Computes the authentication tag.
 *
 * This finishes the authenticated encryption process.
 *
 * @param[in] ctx		- the context.
 * @param[out] tag		- the buffer for the authentication tag.
 * @param[in] tag_len	- the length of the buffer.
 */
#if AC_METHD == CCFBH
#define ac_tag(ctx, tag, tag_len) ac_ccfbh_tag(ctx, tag, tag_len)
#elif AC_METHD == CCM
#define ac_tag(ctx, tag, tag_len) ac_ccm_tag(ctx, tag, tag_len)
#elif AC_METHD == GCM
#define ac_tag(ctx, tag, tag_len) ac_gcm_tag(ctx, tag, tag_len)
#elif AC_METHD == HB2
#define ac_tag(ctx, tag, tag_len) ac_hb2_tag(ctx, tag, tag_len)
#elif AC_METHD == MASHA
#define ac_tag(ctx, tag, tag_len) ac_masha_tag(ctx, tag, tag_len)
#elif AC_METHD == OCB3
#define ac_tag(ctx, tag, tag_len) ac_ocb3_tag(ctx, tag, tag_len)
#endif

void ac_ccfbh_tag(ac_ccfbh_ctx_t ctx, unsigned char *tag, size_t tag_len);

void ac_ccm_tag(ac_ccm_ctx_t ctx, unsigned char *tag, size_t tag_len);

void ac_gcm_tag(ac_gcm_ctx_t ctx, unsigned char *tag, size_t tag_len);

void ac_hb2_tag(ac_hb2_ctx_t ctx, unsigned char *tag, size_t tag_len);

void ac_masha_tag(ac_masha_ctx_t ctx, unsigned char *tag, size_t tag_len);

void ac_ocb3_tag(ac_ocb3_ctx_t ctx, unsigned char *tag, size_t tag_len);

/**
 * Checks the authentication tag.
 *
 * This finishes the decryption-verification process.
 *
 * @param[in] ctx		- the context.
 * @param[in] tag		- the authentication tag received.
 * @param[in] tag_len	- the length of the authentication tag received.
 * @return 1 if the tag is valid, 0 otherwise.
 */
#if AC_METHD == CCFBH
#define ac_check(ctx, tag, tag_len) ac_ccfbh_check(ctx, tag, tag_len)
#elif AC_METHD == CCM
#define ac_check(ctx, tag, tag_len) ac_ccm_check(ctx, tag, tag_len)
#elif AC_METHD == GCM
#define ac_check(ctx, tag, tag_len) ac_gcm_check(ctx, tag, tag_len)
#elif AC_METHD == HB2
#define ac_check(ctx, tag, tag_len) ac_hb2_check(ctx, tag, tag_len)
#elif AC_METHD == MASHA
#define ac_check(ctx, tag, tag_len) ac_masha_check(ctx, tag, tag_len)
#elif AC_METHD == OCB3
#define ac_check(ctx, tag, tag_len) ac_ocb3_check(ctx, tag, tag_len)
#endif

int ac_ccfbh_check(ac_ccfbh_ctx_t ctx, unsigned char *tag, size_t tag_len);

int ac_ccm_check(ac_ccm_ctx_t ctx, unsigned char *tag, size_t tag_len);

int ac_gcm_check(ac_gcm_ctx_t ctx, unsigned char *tag, size_t tag_len);

int ac_hb2_check(ac_hb2_ctx_t ctx, unsigned char *tag, size_t tag_len);

int ac_masha_check(ac_masha_ctx_t ctx, unsigned char *tag, size_t tag_len);

int ac_ocb3_check(ac_ocb3_ctx_t ctx, unsigned char *tag, size_t tag_len);

#endif /* RELIC_AC_H_ */
