/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @defgroup bc Block ciphers.
 */

/**
 * @file
 *
 * Interface of the module for block ciphers.
 *
 * @version $Id$
 * @ingroup bc
 */

#ifndef RELIC_BC_H
#define RELIC_BC_H

#include <stddef.h>

#include "relic_types.h"
#include "relic_bc_conf.h"

/*============================================================================*/
/* Constant definitions                                                       */
/*============================================================================*/

#define BC_AES128_BLOCK_LEN 16
#define BC_AES128_KEY_LEN 16

/**
 * Block length in bytes.
 */
#if BC_METHD == AES128
#define BC_BLOCK_LEN BC_AES128_BLOCK_LEN
#endif

/**
 * Key length in bytes.
 */
#if BC_METHD == AES128
#define BC_BLOCK_LEN BC_AES128_BLOCK_LEN
#define BC_KEY_LEN BC_AES128_KEY_LEN
#endif

/*============================================================================*/
/* Type definitions                                                           */
/*============================================================================*/

/**
 * Represents a block of bytes.
 */
#if defined(__ICC430__)
typedef unsigned char bc_priv_t[BC_BLOCK_LEN];
#define bc_t align bc_priv_t
#else
typedef align unsigned char bc_t[BC_BLOCK_LEN];
#endif

/**
 * Represents a key.
 */
#if defined(__ICC430__)
typedef unsigned char bc_key_priv_t[BC_KEY_LEN];
#define bc_key_t align bc_key_priv_t
#else
typedef align unsigned char bc_key_t[BC_KEY_LEN];
#endif

//Context structs are defined by the backend.
#include "relic_bc_types_low.h"

/**
 * Represents the context of the encryption / decryption process.
 */
#if BC_METHD == AES128
typedef bc_aes128_ctx_st bc_ctx_st;
#endif

/**
 * Pointer to the context structure.
 */
typedef bc_ctx_st bc_ctx_t[1];

/*============================================================================*/
/* Function prototypes                                                        */
/*============================================================================*/

/**
 * Sets the encryption key.
 *
 * After this call, the context becomes ready to carry encryption.
 *
 * @param[in,out] ctx		- the encryption context.
 * @param[in] key			- the key
 * @param[in] key_len		- the key length, in bytes.
 * @throw ERR_NO_VALID		- if the key length is invalid.
 */
#if BC_METHD == AES128
#define bc_enc_key(ctx, key, key_len) bc_aes128_enc_key(ctx, key, key_len)
#endif

void bc_aes128_enc_key(bc_ctx_t ctx, unsigned char *key, size_t key_len);

/**
 * Sets the decryption key.
 *
 * After this call, the context becomes ready to carry decryption.
 *
 * @param[in,out] ctx		- the decryption context.
 * @param[in] key			- the key
 * @param[in] key_len		- the key length, in bytes.
 * @throw ERR_NO_VALID		- if the key length is invalid.
 */
#if BC_METHD == AES128
#define bc_dec_key(ctx, key, key_len) bc_aes128_dec_key(ctx, key, key_len)
#endif

void bc_aes128_dec_key(bc_ctx_t ctx, unsigned char *key, size_t key_len);

/**
 * Starts the encryption of a block.
 *
 * Input and output blocks must be aligned and have BC_BLOCK_LEN bytes.
 *
 * @param[in,out] ctx		- the encryption context.
 * @param[out] output		- the ciphertext generated.
 * @param[in] input			- the plaintext to encrypt.
 */
#if BC_METHD == AES128
#define bc_enc_start(ctx, output, input) bc_aes128_enc_start(ctx, output, input)
#endif

void bc_aes128_enc_start(bc_ctx_t ctx, bc_t output, bc_t input);

/**
 * Finishes the encryption of a block.
 *
 * Output block must be aligned and have BC_BLOCK_LEN bytes.
 *
 * @param[in,out] ctx		- the encryption context.
 * @param[out] output		- the ciphertext generated.
 */
#ifndef BC_SPLIT
#define bc_enc_end(ctx, output)
#elif BC_METHD == AES128
#define bc_enc_end(ctx, output) bc_aes128_enc_end(ctx, output)
#endif

void bc_aes128_enc_end(bc_ctx_t ctx, bc_t output);

/**
 * Encrypts a block.
 *
 * Input and output blocks must be aligned and have BC_BLOCK_LEN bytes.
 *
 * @param[in,out] ctx		- the encryption context.
 * @param[out] output		- the ciphertext generated.
 * @param[in] input			- the plaintext to encrypt.
 */
#ifdef BC_SPLIT
#define bc_enc(ctx, output, input) bc_enc_start(ctx, output, input); bc_enc_end(ctx, output)
#else
#define bc_enc(ctx, output, input) bc_enc_start(ctx, output, input)
#endif

#ifdef BC_SPLIT
#define bc_aes128_enc(ctx, output, input) bc_aes128_enc_start(ctx, output, input); bc_aes128_enc_end(ctx, output)
#else
#define bc_aes128_enc(ctx, output, input) bc_aes128_enc_start(ctx, output, input)
#endif

/**
 * Starts the decryption of a block.
 *
 * Input and output blocks must be aligned and have BC_BLOCK_LEN bytes.
 *
 * @param[in,out] ctx		- the decryption context.
 * @param[out] output		- the plaintext generated.
 * @param[in] input			- the ciphertext to decrypt.
 */
#if BC_METHD == AES128
#define bc_dec_start(ctx, output, input) bc_aes128_dec_start(ctx, output, input)
#endif

void bc_aes128_dec_start(bc_ctx_t ctx, bc_t output, bc_t input);

/**
 * Finishes the decryption of a block.
 *
 * Output block must be aligned and have BC_BLOCK_LEN bytes.
 *
 * @param[in,out] ctx		- the decryption context.
 * @param[out] output		- the plaintext generated.
 */
#ifndef BC_SPLIT
#define bc_dec_end(ctx, output)
#elif BC_METHD == AES128
#define bc_dec_end(ctx, output) bc_aes128_dec_end(ctx, output)
#endif

void bc_aes128_dec_end(bc_ctx_t ctx, bc_t output);

/**
 * Decrypts a block.
 *
 * Input and output blocks must be aligned and have BC_BLOCK_LEN bytes.
 *
 * @param[in,out] ctx		- the encryption context.
 * @param[out] output		- the plaintext generated.
 * @param[in] input			- the ciphertext to decrypt.
 */
#ifdef BC_SPLIT
#define bc_dec(ctx, output, input) bc_dec_start(ctx, output, input); bc_dec_end(ctx, output)
#else
#define bc_dec(ctx, output, input) bc_dec_start(ctx, output, input)
#endif

#ifdef BC_SPLIT
#define bc_aes128_dec(ctx, output, input) bc_aes128_dec_start(ctx, output, input); bc_aes128_dec_end(ctx, output)
#else
#define bc_aes128_dec(ctx, output, input) bc_aes128_dec_start(ctx, output, input)
#endif

/**
 * Initializes the context for encryption / decryption using a mode of
 * operation.
 *
 * @param ctx[in,out]		- the context.
 * @param iv[in]			- the initialization vector (IV).
 * @param iv_len[in]		- the length of the IV in bytes.
 */
void bc_init(bc_ctx_t ctx, unsigned char *iv, size_t iv_len);

/**
 * Encrypts using the CTR mode.
 *
 * The context must be initialized beforehand with bc_init.
 * The input and output buffers must be aligned. They must have BC_BLOCK_LEN
 * bytes, except for the last call when they can have 0 < len <= BC_BLOCK_LEN.
 *
 * @param ctx[in,out]		- the encryption context.
 * @param output[out]		- the ciphertext generated.
 * @param input[in]			- the plaintext to encrypt.
 * @param len[in]			- the length of the input and output in bytes.
 */
void bc_ctr_enc(bc_ctx_t ctx, unsigned char *output, unsigned char *input, size_t len);

/**
 * Decrypts using the CTR mode.
 *
 * The context must be initialized beforehand with bc_init.
 * The input and output buffers must be aligned. They must have BC_BLOCK_LEN
 * bytes, except for the last call when they can have 0 < len <= BC_BLOCK_LEN.
 *
 * @param ctx[in,out]		- the decryption context.
 * @param output[out]		- the plaintext generated.
 * @param input[in]			- the ciphertext to decrypt.
 * @param len[in]			- the length of the input and output in bytes.
 */
#define bc_ctr_dec(ctx, output, input, len) bc_ctr_enc(ctx, output, input, len)

#endif /* !RELIC_BC_H */
