/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of the CCFBH authenticated cipher functions.
 *
 * @version $Id$
 * @ingroup ac
 */
#include <string.h>

#include "relic_ac.h"
#include "relic_ac_low.h"
#include "relic_bc_low.h"
#include "relic_core.h"
#include "relic_util.h"

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

void ac_ccfbh_key(ac_ccfbh_ctx_t ctx, unsigned char *key, size_t key_len) {
	if (key_len != BC_KEY_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}
	bc_enc_key(ctx->bc_ctx, key, key_len);
}

void ac_ccfbh_init(ac_ccfbh_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len) {
	ctx->i = (unsigned int *) (ctx->last_c + AC_CCFBH_BLOCK_LEN);

	if (key_len != BC_KEY_LEN || iv_len != AC_CCFBH_IV_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}

	ac_ccfbh_copy_low(ctx->last_c, iv);
	*(ctx->i) = 1;
	memset(ctx->last_c + AC_CCFBH_BLOCK_LEN + sizeof(unsigned), 0,
			BC_BLOCK_LEN - AC_CCFBH_BLOCK_LEN - sizeof(unsigned));
	memset(ctx->tag, 0, sizeof(ctx->tag));
}

void ac_ccfbh_data(ac_ccfbh_ctx_t ctx, unsigned char *data, size_t data_len) {
	THROW(ERR_NO_VALID);
}

void ac_ccfbh_enc(ac_ccfbh_ctx_t ctx, unsigned char *output,
		unsigned char *input, size_t input_len) {
	bc_t t;

	bc_enc_start(ctx->bc_ctx, t, ctx->last_c);
	(*(ctx->i))++;
	bc_enc_end(ctx->bc_ctx, t);

	if (input_len < AC_CCFBH_BLOCK_LEN) {
		bc_nxor_low(ctx->last_c, t, input, input_len);
		memcpy(output, ctx->last_c, input_len);
		(*(ctx->i))++;
		memcpy(ctx->last_c + input_len, t + input_len,
				AC_CCFBH_BLOCK_LEN - input_len);
		ctx->last_c[input_len] ^= 1;
	} else {
		ac_ccfbh_xor_low(ctx->last_c, t, input);
		ac_ccfbh_copy_low(output, ctx->last_c);
	}

	ac_ccfbh_xortag_low(ctx->tag, t + AC_CCFBH_BLOCK_LEN);
}

void ac_ccfbh_dec(ac_ccfbh_ctx_t ctx, unsigned char *output,
		unsigned char *input, size_t input_len) {
	bc_t t;

	bc_enc_start(ctx->bc_ctx, t, ctx->last_c);

	if (input_len < AC_CCFBH_BLOCK_LEN) {
		memcpy(ctx->last_c, input, input_len);
		bc_enc_end(ctx->bc_ctx, t);
		bc_nxor_low(output, t, input, input_len);
		(*(ctx->i)) += 2;
		memcpy(ctx->last_c + input_len, t + input_len,
				AC_CCFBH_BLOCK_LEN - input_len);
		ctx->last_c[input_len] ^= 1;
	} else {
		ac_ccfbh_copy_low(ctx->last_c, input);
		(*(ctx->i))++;
		bc_enc_end(ctx->bc_ctx, t);

		ac_ccfbh_xor_low(output, t, input);
	}

	ac_ccfbh_xortag_low(ctx->tag, t + AC_CCFBH_BLOCK_LEN);
}

void ac_ccfbh_tag(ac_ccfbh_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	if (tag_len != AC_CCFBH_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}

	memset(ctx->last_c + AC_CCFBH_BLOCK_LEN + sizeof(ctx->i), 0,
			BC_BLOCK_LEN - AC_CCFBH_BLOCK_LEN - sizeof(unsigned));

	bc_enc_start(ctx->bc_ctx, ctx->last_c, ctx->last_c);
	bc_enc_end(ctx->bc_ctx, ctx->last_c);

	ac_ccfbh_xortag_low(ctx->tag, ctx->last_c + AC_CCFBH_BLOCK_LEN);
	memcpy(tag, ctx->tag, AC_CCFBH_TAG_LEN);
}

int ac_ccfbh_check(ac_ccfbh_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	unsigned char computed_tag[AC_CCFBH_TAG_LEN];

	if (tag_len != AC_CCFBH_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return 0;
	}

	ac_ccfbh_tag(ctx, computed_tag, sizeof(computed_tag));

	if (util_cmp_const(tag, computed_tag, AC_CCFBH_TAG_LEN) == CMP_EQ) {
		return 1;
	} else {
		return 0;
	}
}
