/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of the CCM authenticated cipher functions.
 *
 * @version $Id$
 * @ingroup ac
 */
#include <stdlib.h>
#include <string.h>

#include "relic_core.h"
#include "relic_ac.h"
#include "relic_ac_low.h"
#include "relic_bc_low.h"
#include "relic_types.h"
#include "relic_util.h"

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

void ac_ccm_key(ac_ccm_ctx_t ctx, unsigned char *key, size_t key_len) {
	if (key_len != BC_KEY_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}
	bc_enc_key(ctx->bc_ctx, key, key_len);
}

void ac_ccm_init(ac_ccm_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len) {
	bc_t t;
	size_t q = 15 - AC_CCM_IV_LEN;
	ctx->data_buf_len = 0;

	if (key_len != BC_KEY_LEN || iv_len != AC_CCM_IV_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}

	//Block 0
	t[0] = (data_len ? (1 << 6) : 0) | (((AC_CCM_TAG_LEN - 2) / 2) << 3)
			| (q - 1);
	memcpy(t + 1, iv, AC_CCM_IV_LEN);
#if (15 - AC_CCM_IV_LEN) > 4
#error "AC_CCM_IV_LEN value not supported"
#endif
	uint32_t Q = util_conv_big(msg_len);
	unsigned char *p = (unsigned char *) &Q;
	memcpy(t + 1 + AC_CCM_IV_LEN, p + (sizeof(uint32_t) - q), q);

	bc_enc_start(ctx->bc_ctx, ctx->last_y, t);

	//Begin block 1
	if (data_len > 0) {
		uint32_t a = util_conv_big(data_len);
		p = (unsigned char *) &a;
		if (data_len < 0xFF00) {
			memcpy(ctx->data_buf, p + 2, 2);
			ctx->data_buf_len += 2;
		} else {
			ctx->data_buf[0] = 0xFF;
			ctx->data_buf[1] = 0xFE;
			memcpy(ctx->data_buf + 2, p, 4);
			ctx->data_buf_len += 4;
		}
	}

	bc_enc_end(ctx->bc_ctx, ctx->last_y);

	//Setup counter
	ctx->bc_ctx->iv[0] = q - 1;
	memcpy(ctx->bc_ctx->iv + 1, iv, AC_CCM_IV_LEN);
	memset(ctx->bc_ctx->iv + 1 + AC_CCM_IV_LEN, 0,
			BC_BLOCK_LEN - 1 - AC_CCM_IV_LEN);
	bc_enc_start(ctx->bc_ctx, ctx->s0, ctx->bc_ctx->iv);
	bc_inc32_low(ctx->bc_ctx->iv);
	bc_enc_end(ctx->bc_ctx, ctx->s0);
}

void ac_ccm_data(ac_ccm_ctx_t ctx, unsigned char *data, size_t data_len) {
	size_t len;

	if (data_len == 0 && ctx->data_buf_len > 0) {
		memset(ctx->data_buf + ctx->data_buf_len, 0,
				BC_BLOCK_LEN - ctx->data_buf_len);
		ctx->data_buf_len = 0;
		bc_xor_low(ctx->last_y, ctx->last_y, ctx->data_buf);
		bc_enc_start(ctx->bc_ctx, ctx->last_y, ctx->last_y);
	}

	while (data_len) {
		len = MIN(BC_BLOCK_LEN - ctx->data_buf_len, data_len);
		memcpy(ctx->data_buf + ctx->data_buf_len, data, len);
		ctx->data_buf_len += len;
		data += len;
		data_len -= len;
		if (ctx->data_buf_len == BC_BLOCK_LEN) {
			ctx->data_buf_len = 0;
			bc_xor_low(ctx->last_y, ctx->last_y, ctx->data_buf);
			bc_enc_start(ctx->bc_ctx, ctx->last_y, ctx->last_y);
		}
	}
}

void ac_ccm_enc(ac_ccm_ctx_t ctx, unsigned char *output, unsigned char *input,
		size_t input_len) {
	bc_t m;
	bc_t t;

	if (input_len < BC_BLOCK_LEN) {
		bc_enc_start(ctx->bc_ctx, t, ctx->bc_ctx->iv);
		memcpy(m, input, input_len);
		memset(m + input_len, 0, BC_BLOCK_LEN - input_len);
		bc_xor_low(ctx->last_y, ctx->last_y, m);
		bc_enc_end(ctx->bc_ctx, t);bc_enc_start(ctx->bc_ctx, ctx->last_y,
				ctx->last_y);
		bc_xor_low(t, m, t);
		memcpy(output, t, input_len);
		bc_enc_end(ctx->bc_ctx, ctx->last_y);
	} else {
		bc_enc_start(ctx->bc_ctx, t, ctx->bc_ctx->iv);
		bc_inc32_low(ctx->bc_ctx->iv);
		bc_xor_low(ctx->last_y, ctx->last_y, input);
		bc_enc_end(ctx->bc_ctx, t);bc_enc_start(ctx->bc_ctx, ctx->last_y,
				ctx->last_y);
		bc_xor_low(output, input, t);
		bc_enc_end(ctx->bc_ctx, ctx->last_y);
	}
}

void ac_ccm_dec(ac_ccm_ctx_t ctx, unsigned char *output, unsigned char *input,
		size_t input_len) {
	bc_t c;
	bc_t t;

	if (input_len < BC_BLOCK_LEN) {
		bc_enc_start(ctx->bc_ctx, t, ctx->bc_ctx->iv);
		memcpy(c, input, input_len);
		bc_enc_end(ctx->bc_ctx, t);
		bc_xor_low(t, c, t);
		memset(t + input_len, 0, BC_BLOCK_LEN - input_len);
		bc_xor_low(ctx->last_y, ctx->last_y, t);
		bc_enc_start(ctx->bc_ctx, ctx->last_y, ctx->last_y);
		memcpy(output, t, input_len);
		bc_enc_end(ctx->bc_ctx, ctx->last_y);
	} else {
		bc_enc_start(ctx->bc_ctx, t, ctx->bc_ctx->iv);
		bc_inc32_low(ctx->bc_ctx->iv);
		bc_enc_end(ctx->bc_ctx, t);
		bc_xor_low(output, input, t);
		bc_xor_low(ctx->last_y, ctx->last_y, output);
		bc_enc_start(ctx->bc_ctx, ctx->last_y, ctx->last_y);
	}
}

void ac_ccm_tag(ac_ccm_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	if (tag_len != AC_CCM_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}
	bc_nxor_low(tag, ctx->last_y, ctx->s0, AC_CCM_TAG_LEN);
}

int ac_ccm_check(ac_ccm_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	unsigned char computed_tag[AC_CCM_TAG_LEN];

	if (tag_len != AC_CCM_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return 0;
	}

	ac_ccm_tag(ctx, computed_tag, sizeof(computed_tag));

	if (util_cmp_const(tag, computed_tag, AC_CCM_TAG_LEN) == CMP_EQ) {
		return 1;
	} else {
		return 0;
	}
}
