/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of the GCM authenticated cipher functions.
 *
 * @version $Id$
 * @ingroup ac
 */

#include <stdlib.h>
#include <string.h>

#include "relic_core.h"
#include "relic_types.h"
#include "relic_ac.h"
#include "relic_ac_low.h"
#include "relic_bc_low.h"
#include "relic_util.h"

/*============================================================================*/
/* Private definitions                                                        */
/*============================================================================*/

#ifndef AC_GCM_REFLC
/** Table used to reflect the bits inside a byte. */
static const unsigned char byte_table[] = { 0x0, 0x80, 0x40, 0xc0, 0x20,
		0xa0, 0x60, 0xe0, 0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70,
		0xf0, 0x8, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8, 0x18,
		0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8, 0x4, 0x84, 0x44,
		0xc4, 0x24, 0xa4, 0x64, 0xe4, 0x14, 0x94, 0x54, 0xd4, 0x34,
		0xb4, 0x74, 0xf4, 0xc, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c,
		0xec, 0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc, 0x2,
		0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2, 0x12, 0x92, 0x52,
		0xd2, 0x32, 0xb2, 0x72, 0xf2, 0xa, 0x8a, 0x4a, 0xca, 0x2a,
		0xaa, 0x6a, 0xea, 0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a,
		0xfa, 0x6, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6, 0x16,
		0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6, 0xe, 0x8e, 0x4e,
		0xce, 0x2e, 0xae, 0x6e, 0xee, 0x1e, 0x9e, 0x5e, 0xde, 0x3e,
		0xbe, 0x7e, 0xfe, 0x1, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61,
		0xe1, 0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1, 0x9,
		0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9, 0x19, 0x99, 0x59,
		0xd9, 0x39, 0xb9, 0x79, 0xf9, 0x5, 0x85, 0x45, 0xc5, 0x25,
		0xa5, 0x65, 0xe5, 0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75,
		0xf5, 0xd, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed, 0x1d,
		0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd, 0x3, 0x83, 0x43,
		0xc3, 0x23, 0xa3, 0x63, 0xe3, 0x13, 0x93, 0x53, 0xd3, 0x33,
		0xb3, 0x73, 0xf3, 0xb, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b,
		0xeb, 0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb, 0x7,
		0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7, 0x17, 0x97, 0x57,
		0xd7, 0x37, 0xb7, 0x77, 0xf7, 0xf, 0x8f, 0x4f, 0xcf, 0x2f,
		0xaf, 0x6f, 0xef, 0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f,
		0xff };
#endif

/**
 * Convert byte stream to GCM representation.
 *
 * Since GCM treats the bits inside a byte in reverse order, we must convert it
 * to a suitable representation.
 *
 * @param[out] c			- the output block.
 * @param[in] a				- the input block.
 */
#ifdef AC_GCM_SGCM
#define convert(C, A) memcpy(C, A, BC_BLOCK_LEN)
#elif defined(AC_GCM_REFLC)
static void convert(unsigned char *c, unsigned char *a) {
	bc_t t;
	int i, j;
#ifdef BIGED
	dig_t *ad = (dig_t *) a;
	dig_t *td = (dig_t *) t;
	for (i = 0; i < AC_GCM_BLOCK_LEN / sizeof(dig_t); i += 1) {
		td[AC_GCM_BLOCK_LEN / sizeof(dig_t) - i - 1] = ad[i];
	}
#else
	for (i = 0; i < AC_GCM_BLOCK_LEN; i += sizeof(dig_t)) {
		for (j = 0; j < sizeof(dig_t); j++) {
			t[(AC_GCM_BLOCK_LEN - i - sizeof(dig_t)) + j] = a[i + (sizeof(dig_t) - j - 1)];
		}
	}
#endif
	memcpy(c, t, AC_GCM_BLOCK_LEN);
}
#else
#if 1
void ac_gcm_convert_low(unsigned char *c, unsigned char *a);
#define convert ac_gcm_convert_low
#else
static void convert(unsigned char *c, unsigned char *a) {
	int i;
#ifdef BIGED
	uint32_t *p = (uint32_t *) block;
	for (i = 0; i < BC_BLOCK_LEN / sizeof(uint32_t); i++) {
		p[i] = util_conv_little(p[i]);
	}
#endif
	for (i = 0; i < BC_BLOCK_LEN; i++) {
		c[i] = byte_table[a[i]];
	}
}
#endif
#endif

/**
 * Inputs a block into the GHASH function.
 *
 * @param[in,out] ctx		- the context.
 * @param[in] input			- the input block.
 */
static void ghash_input(ac_gcm_ctx_t ctx, bc_t input) {
	bc_t t;
	//xor (field addition)
	convert(t, input);
	bc_xor_low(ctx->last_y, ctx->last_y, t);
	//binary field multiplication
	ac_gcm_mul_low((dig_t *) ctx->last_y, (dig_t *) ctx->last_y,
			(dig_t *) ctx->table);
}

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

void ac_gcm_key(ac_gcm_ctx_t ctx, unsigned char *key, size_t key_len) {
	bc_t h;

	ctx->len_a = ctx->len_c = 0;

	if (key_len != BC_KEY_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}

	memset(ctx->last_y, 0, BC_BLOCK_LEN);
	bc_enc_key(ctx->bc_ctx, key, key_len);

	//H = CIPH_K(0^128)
	bc_enc(ctx->bc_ctx, h, ctx->last_y);
	convert(h, h);
	ac_gcm_tab_low(ctx->table, h);
}

void ac_gcm_init(ac_gcm_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len) {
	if (key_len != BC_KEY_LEN || iv_len != AC_GCM_IV_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}
	//J_0 = IV || 0^31 || 1
	memcpy(ctx->bc_ctx->iv, iv, AC_GCM_IV_LEN);
	memset(ctx->bc_ctx->iv + AC_GCM_IV_LEN, 0, BC_BLOCK_LEN - AC_GCM_IV_LEN);
	ctx->bc_ctx->iv[BC_BLOCK_LEN - 1] = 1;
	bc_inc32_low(ctx->bc_ctx->iv);
}

void ac_gcm_data(ac_gcm_ctx_t ctx, unsigned char *data, size_t data_len) {
	unsigned char tmp[BC_BLOCK_LEN];
	unsigned char *t = data;
	if (data_len < BC_BLOCK_LEN) {
		t = tmp;
		memcpy(tmp, data, data_len);
		memset(tmp + data_len, 0, BC_BLOCK_LEN - data_len);
	}
	ghash_input(ctx, t);
	ctx->len_a += data_len;
}

void ac_gcm_enc_sub_low(unsigned char *output, unsigned char *input, bc_t t, bc_t last_y);

void ac_gcm_enc(ac_gcm_ctx_t ctx, unsigned char *output, unsigned char *input,
		size_t input_len) {
	bc_t t;

	bc_enc_start(ctx->bc_ctx, t, ctx->bc_ctx->iv);
	bc_inc32_low(ctx->bc_ctx->iv);
	bc_enc_end(ctx->bc_ctx, t);
	if (input_len < BC_BLOCK_LEN) {
		bc_nxor_low(output, input, t, input_len);
		memset(t, 0, sizeof(t));
		memcpy(t, output, input_len);
		convert(t, t);
		bc_xor_low(ctx->last_y, ctx->last_y, t);
	} else {
//		ac_gcm_enc_sub_low(output, input, t, ctx->last_y);
		bc_xor_low(output, input, t);
		convert(t, output);
		bc_xor_low(ctx->last_y, ctx->last_y, t);
	}

	ac_gcm_mul_low((dig_t *) ctx->last_y, (dig_t *) ctx->last_y,
			(dig_t *) ctx->table);

	ctx->len_c += input_len;
}

void ac_gcm_dec(ac_gcm_ctx_t ctx, unsigned char *output, unsigned char *input,
		size_t input_len) {
	bc_t t;

	bc_enc_start(ctx->bc_ctx, t, ctx->bc_ctx->iv);
	bc_inc32_low(ctx->bc_ctx->iv);
	bc_enc_end(ctx->bc_ctx, t);
	if (input_len < BC_BLOCK_LEN) {
		bc_nxor_low(output, input, t, input_len);
		memset(t, 0, sizeof(t));
		memcpy(t, input, input_len);
		convert(t, t);
		bc_xor_low(ctx->last_y, ctx->last_y, t);
	} else {
		bc_xor_low(output, input, t);
		convert(t, input);
		bc_xor_low(ctx->last_y, ctx->last_y, t);
	}
	ac_gcm_mul_low((dig_t *) ctx->last_y, (dig_t *) ctx->last_y,
			(dig_t *) ctx->table);
	ctx->len_c += input_len;
}

void ac_gcm_tag(ac_gcm_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	bc_t t;

	if (tag_len != AC_GCM_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}

	//Build [len(A)]_64 || [len(C)]_64
	memset(t, 0, sizeof(t));
	uint32_t len_a = util_conv_big(8 * ctx->len_a);
	uint32_t len_c = util_conv_big(8 * ctx->len_c);
	memcpy(t + (BC_BLOCK_LEN / 2) - sizeof(uint32_t), &len_a, sizeof(uint32_t));
	memcpy(t + BC_BLOCK_LEN - sizeof(uint32_t), &len_c, sizeof(uint32_t));
	ghash_input(ctx, t);

	//Compute S
	convert(ctx->last_y, ctx->last_y);

	//Compute GCTR_K(J_0, S)
	memset(ctx->bc_ctx->iv + AC_GCM_IV_LEN, 0, BC_BLOCK_LEN - AC_GCM_IV_LEN);
	ctx->bc_ctx->iv[BC_BLOCK_LEN - 1] = 1;
	bc_enc_start(ctx->bc_ctx, t, ctx->bc_ctx->iv);
	bc_enc_end(ctx->bc_ctx, t);
	bc_xor_low(t, t, ctx->last_y);

	memcpy(tag, t, AC_GCM_TAG_LEN);
}

int ac_gcm_check(ac_gcm_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	unsigned char computed_tag[AC_GCM_TAG_LEN];

	if (tag_len != AC_GCM_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return 0;
	}

	ac_gcm_tag(ctx, computed_tag, sizeof(computed_tag));

	if (util_cmp_const(tag, computed_tag, AC_GCM_TAG_LEN) == CMP_EQ) {
		return 1;
	} else {
		return 0;
	}
}
