/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of the Hummingbird-2 authenticated cipher functions.
 *
 * @version $Id$
 * @ingroup fp
 */

#include <stdio.h>
#include <string.h>

#include "relic_ac.h"
#include "relic_ac_low.h"
#include "relic_types.h"
#include "relic_core.h"
#include "relic_test.h"
#include "relic_util.h"

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

void ac_hb2_key(ac_hb2_ctx_t ctx, unsigned char *key, size_t key_len) {
}

#ifndef ac_hb2_init
void ac_hb2_init(ac_hb2_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len) {
	uint16_t t1, t2, t3, t4, t;

	memcpy(ctx->K, key, sizeof(ctx->K));
	memcpy(ctx->IV, iv, sizeof(ctx->IV));

	ctx->R[0] = ctx->IV[0];
	ctx->R[1] = ctx->IV[1];
	ctx->R[2] = ctx->IV[2];
	ctx->R[3] = ctx->IV[3];
	ctx->R[4] = ctx->IV[0];
	ctx->R[5] = ctx->IV[1];
	ctx->R[6] = ctx->IV[2];
	ctx->R[7] = ctx->IV[3];

	for (uint16_t i = 0; i < 4; i++) {
		t1 = ac_hb2_wd16_low((ctx->R[0] + i) ^ ctx->K[0], ctx->K[1], ctx->K[2], ctx->K[3]);
		t2 = ac_hb2_wd16_low((ctx->R[1] + t1) ^ ctx->K[4], ctx->K[5], ctx->K[6], ctx->K[7]);
		t3 = ac_hb2_wd16_low((ctx->R[2] + t2) ^ ctx->K[0], ctx->K[1], ctx->K[2], ctx->K[3]);
		t4 = ac_hb2_wd16_low((ctx->R[3] + t3) ^ ctx->K[4], ctx->K[5], ctx->K[6], ctx->K[7]);
		t = (ctx->R[0] + t4);
		ctx->R[0] = (t << 3) | (t >> 13);
		t = (ctx->R[1] + t1);
		ctx->R[1] = (t >> 1) | (t << 15);
		t = (ctx->R[2] + t2);
		ctx->R[2] = (t << 8) | (t >> 8);
		t = (ctx->R[3] + t3);
		ctx->R[3] = (t << 1) | (t >> 15);
		ctx->R[4] = ctx->R[4] ^ ctx->R[0];
		ctx->R[5] = ctx->R[5] ^ ctx->R[1];
		ctx->R[6] = ctx->R[6] ^ ctx->R[2];
		ctx->R[7] = ctx->R[7] ^ ctx->R[3];
	}
}
#endif

void ac_hb2_data(ac_hb2_ctx_t ctx, unsigned char *data, size_t data_len) {
	uint16_t t;
	ac_hb2_enc(ctx, (unsigned char *) &t, data, data_len);
}

#ifndef ac_hb2_enc
void ac_hb2_enc(ac_hb2_ctx_t ctx, unsigned char *output, unsigned char *input,
		size_t input_len) {
	uint16_t t1, t2, t3, c;
	uint16_t p = *((uint16_t *) input);

	t1 = ac_hb2_wd16_low((ctx->R[0] + p) ^ ctx->K[0], ctx->K[1], ctx->K[2],
			ctx->K[3]);
	t2 = ac_hb2_wd16_low((ctx->R[1] + t1) ^ ctx->K[4] ^ ctx->R[4],
			ctx->K[5] ^ ctx->R[5], ctx->K[6] ^ ctx->R[6],
			ctx->K[7] ^ ctx->R[7]);
	t3 = ac_hb2_wd16_low((ctx->R[2] + t2) ^ ctx->K[0] ^ ctx->R[4],
			ctx->K[1] ^ ctx->R[5], ctx->K[2] ^ ctx->R[6],
			ctx->K[3] ^ ctx->R[7]);
	c = ac_hb2_wd16_low((ctx->R[3] + t3) ^ ctx->K[4], ctx->K[5], ctx->K[6],
			ctx->K[7]) + ctx->R[0];

	*((uint16_t *) output) = c;

	ctx->R[0] = ctx->R[0] + t3;
	ctx->R[1] = ctx->R[1] + t1;
	ctx->R[2] = ctx->R[2] + t2;
	ctx->R[3] = ctx->R[3] + ctx->R[0] + t1;
	ctx->R[4] = ctx->R[4] ^ ctx->R[0];
	ctx->R[5] = ctx->R[5] ^ ctx->R[1];
	ctx->R[6] = ctx->R[6] ^ ctx->R[2];
	ctx->R[7] = ctx->R[7] ^ ctx->R[3];
}
#endif

#ifndef ac_hb2_dec
void ac_hb2_dec(ac_hb2_ctx_t ctx, unsigned char *msg, unsigned char *cipher_msg,
		size_t cipher_msg_len) {
	uint16_t t1, t2, t3, p;
	uint16_t c = *((uint16_t *) cipher_msg);

	t3 = (ac_hb2_wd16inv_low(c - ctx->R[0], ctx->K[5], ctx->K[6], ctx->K[7])
			^ ctx->K[4]) - ctx->R[3];
	t2 = (ac_hb2_wd16inv_low(t3, ctx->K[1] ^ ctx->R[5], ctx->K[2] ^ ctx->R[6],
			ctx->K[3] ^ ctx->R[7]) ^ (ctx->K[0] ^ ctx->R[4])) - ctx->R[2];
	t1 = (ac_hb2_wd16inv_low(t2, ctx->K[5] ^ ctx->R[5], ctx->K[6] ^ ctx->R[6],
			ctx->K[7] ^ ctx->R[7]) ^ (ctx->K[4] ^ ctx->R[4])) - ctx->R[1];
	p = (ac_hb2_wd16inv_low(t1, ctx->K[1], ctx->K[2], ctx->K[3]) ^ ctx->K[0])
			- ctx->R[0];

	*((uint16_t *) msg) = p;

	ctx->R[0] = ctx->R[0] + t3;
	ctx->R[1] = ctx->R[1] + t1;
	ctx->R[2] = ctx->R[2] + t2;
	ctx->R[3] = ctx->R[3] + ctx->R[0] + t1;
	ctx->R[4] = ctx->R[4] ^ ctx->R[0];
	ctx->R[5] = ctx->R[5] ^ ctx->R[1];
	ctx->R[6] = ctx->R[6] ^ ctx->R[2];
	ctx->R[7] = ctx->R[7] ^ ctx->R[3];
}
#endif

void ac_hb2_tag(ac_hb2_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	uint16_t *tagd = (uint16_t *) tag;
	uint16_t t;
	uint16_t m;

	if (tag_len != AC_HB2_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}

	m = ctx->IV[0] + ctx->R[0] + ctx->R[2] + (AC_HB2_TAG_LEN / 2);
	ac_hb2_enc(ctx, (unsigned char *) &t, (unsigned char *) &m, 2);
	m = ctx->IV[1] + ctx->R[0] + ctx->R[2];
	ac_hb2_enc(ctx, (unsigned char *) &t, (unsigned char *) &m, 2);
	m = ctx->IV[2] + ctx->R[0] + ctx->R[2];
	ac_hb2_enc(ctx, (unsigned char *) &t, (unsigned char *) &m, 2);
	m = ctx->IV[3] + ctx->R[0] + ctx->R[2];
	ac_hb2_enc(ctx, (unsigned char *) &t, (unsigned char *) &m, 2);
	tagd[0] = t;
	for (int i = 1; i < AC_HB2_TAG_LEN / 2; i++) {
		m = ctx->R[0] + ctx->R[2];
		ac_hb2_enc(ctx, (unsigned char *) &t, (unsigned char *) &m, 2);
		tagd[i] = t;
	}
}

int ac_hb2_check(ac_hb2_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	unsigned char computed_tag[AC_HB2_TAG_LEN];

	if (tag_len != AC_HB2_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return 0;
	}

	ac_hb2_tag(ctx, computed_tag, sizeof(computed_tag));

	if (util_cmp_const(tag, computed_tag, AC_HB2_TAG_LEN) == CMP_EQ) {
		return 1;
	} else {
		return 0;
	}
}
