/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of the MASHA authenticated cipher functions.
 *
 * @version $Id$
 * @ingroup ac
 */

#include <stdio.h>
#include <string.h>

#include "relic_core.h"
#include "relic_types.h"
#include "relic_ac.h"
#include "relic_util.h"

/*============================================================================*/
/* Private definitions                                                        */
/*============================================================================*/

/* In MASHA, a 32-bit word is viewed as a GF((2^8)^4) element using an extension
 * tower. The upper extension has modulus a^4 + p*a^3 + q*a^2 + r*a + s for some
 * p, q, r, s in GF(2^8). There are two such towers, with different p, q, r, s
 * values.
 * MASHA uses multiplication by 'a'. That is, a 32-bit value viewed as the
 * field element v_3*a^3 + v_2*a^2 + v_1*a + v_0 is multiplied by a, resulting
 * in v_3*a^4 + v_2*a^3 + v_1*a^2 + v_0*a (each v_i is a GF(2^8) element, i.e. a
 * byte). But a^4 = p*a^3 + q*a^2 + r*a + s (due to the modulus), therefore we
 * can write the result as
 * v_3*(p*a^3 + q*a^2 + r*a + s) + v_2*a^3 + v_1*a^2 + v_0*a.
 * This can be computed by shifting the three lower bytes and multiplying v_3
 * by (p*a^3 + q*a^2 + r*a + s). The tables below provide the result of this
 * multiplication for every byte value, for the two towers used by MASHA.
 */

/**
 * Lookup table used in the multiplication by alpha_0.
 */
static const uint32_t mul_alpha0[256] = {
	0x00000000, 0x9CA7D3E6, 0x91E70F65, 0x0D40DC83, 0x8B671ECA, 0x17C0CD2C, 0x1A8011AF, 0x8627C249,
	0xBFCE3C3D, 0x2369EFDB, 0x2E293358, 0xB28EE0BE, 0x34A922F7, 0xA80EF111, 0xA54E2D92, 0x39E9FE74,
	0xD735787A, 0x4B92AB9C, 0x46D2771F, 0xDA75A4F9, 0x5C5266B0, 0xC0F5B556, 0xCDB569D5, 0x5112BA33,
	0x68FB4447, 0xF45C97A1, 0xF91C4B22, 0x65BB98C4, 0xE39C5A8D, 0x7F3B896B, 0x727B55E8, 0xEEDC860E,
	0x076AF0F4, 0x9BCD2312, 0x968DFF91, 0x0A2A2C77, 0x8C0DEE3E, 0x10AA3DD8, 0x1DEAE15B, 0x814D32BD,
	0xB8A4CCC9, 0x24031F2F, 0x2943C3AC, 0xB5E4104A, 0x33C3D203, 0xAF6401E5, 0xA224DD66, 0x3E830E80,
	0xD05F888E, 0x4CF85B68, 0x41B887EB, 0xDD1F540D, 0x5B389644, 0xC79F45A2, 0xCADF9921, 0x56784AC7,
	0x6F91B4B3, 0xF3366755, 0xFE76BBD6, 0x62D16830, 0xE4F6AA79, 0x7851799F, 0x7511A51C, 0xE9B676FA,
	0x0ED44941, 0x92739AA7, 0x9F334624, 0x039495C2, 0x85B3578B, 0x1914846D, 0x145458EE, 0x88F38B08,
	0xB11A757C, 0x2DBDA69A, 0x20FD7A19, 0xBC5AA9FF, 0x3A7D6BB6, 0xA6DAB850, 0xAB9A64D3, 0x373DB735,
	0xD9E1313B, 0x4546E2DD, 0x48063E5E, 0xD4A1EDB8, 0x52862FF1, 0xCE21FC17, 0xC3612094, 0x5FC6F372,
	0x662F0D06, 0xFA88DEE0, 0xF7C80263, 0x6B6FD185, 0xED4813CC, 0x71EFC02A, 0x7CAF1CA9, 0xE008CF4F,
	0x09BEB9B5, 0x95196A53, 0x9859B6D0, 0x04FE6536, 0x82D9A77F, 0x1E7E7499, 0x133EA81A, 0x8F997BFC,
	0xB6708588, 0x2AD7566E, 0x27978AED, 0xBB30590B, 0x3D179B42, 0xA1B048A4, 0xACF09427, 0x305747C1,
	0xDE8BC1CF, 0x422C1229, 0x4F6CCEAA, 0xD3CB1D4C, 0x55ECDF05, 0xC94B0CE3, 0xC40BD060, 0x58AC0386,
	0x6145FDF2, 0xFDE22E14, 0xF0A2F297, 0x6C052171, 0xEA22E338, 0x768530DE, 0x7BC5EC5D, 0xE7623FBB,
	0x1C019282, 0x80A64164, 0x8DE69DE7, 0x11414E01, 0x97668C48, 0x0BC15FAE, 0x0681832D, 0x9A2650CB,
	0xA3CFAEBF, 0x3F687D59, 0x3228A1DA, 0xAE8F723C, 0x28A8B075, 0xB40F6393, 0xB94FBF10, 0x25E86CF6,
	0xCB34EAF8, 0x5793391E, 0x5AD3E59D, 0xC674367B, 0x4053F432, 0xDCF427D4, 0xD1B4FB57, 0x4D1328B1,
	0x74FAD6C5, 0xE85D0523, 0xE51DD9A0, 0x79BA0A46, 0xFF9DC80F, 0x633A1BE9, 0x6E7AC76A, 0xF2DD148C,
	0x1B6B6276, 0x87CCB190, 0x8A8C6D13, 0x162BBEF5, 0x900C7CBC, 0x0CABAF5A, 0x01EB73D9, 0x9D4CA03F,
	0xA4A55E4B, 0x38028DAD, 0x3542512E, 0xA9E582C8, 0x2FC24081, 0xB3659367, 0xBE254FE4, 0x22829C02,
	0xCC5E1A0C, 0x50F9C9EA, 0x5DB91569, 0xC11EC68F, 0x473904C6, 0xDB9ED720, 0xD6DE0BA3, 0x4A79D845,
	0x73902631, 0xEF37F5D7, 0xE2772954, 0x7ED0FAB2, 0xF8F738FB, 0x6450EB1D, 0x6910379E, 0xF5B7E478,
	0x12D5DBC3, 0x8E720825, 0x8332D4A6, 0x1F950740, 0x99B2C509, 0x051516EF, 0x0855CA6C, 0x94F2198A,
	0xAD1BE7FE, 0x31BC3418, 0x3CFCE89B, 0xA05B3B7D, 0x267CF934, 0xBADB2AD2, 0xB79BF651, 0x2B3C25B7,
	0xC5E0A3B9, 0x5947705F, 0x5407ACDC, 0xC8A07F3A, 0x4E87BD73, 0xD2206E95, 0xDF60B216, 0x43C761F0,
	0x7A2E9F84, 0xE6894C62, 0xEBC990E1, 0x776E4307, 0xF149814E, 0x6DEE52A8, 0x60AE8E2B, 0xFC095DCD,
	0x15BF2B37, 0x8918F8D1, 0x84582452, 0x18FFF7B4, 0x9ED835FD, 0x027FE61B, 0x0F3F3A98, 0x9398E97E,
	0xAA71170A, 0x36D6C4EC, 0x3B96186F, 0xA731CB89, 0x211609C0, 0xBDB1DA26, 0xB0F106A5, 0x2C56D543,
	0xC28A534D, 0x5E2D80AB, 0x536D5C28, 0xCFCA8FCE, 0x49ED4D87, 0xD54A9E61, 0xD80A42E2, 0x44AD9104,
	0x7D446F70, 0xE1E3BC96, 0xECA36015, 0x7004B3F3, 0xF62371BA, 0x6A84A25C, 0x67C47EDF, 0xFB63AD39,
};

/**
 * Lookup table used in the multiplication by alpha_1.
 */
static const uint32_t mul_alpha1[256] = {
	0x00000000, 0x4DA6615C, 0x9A01C2B8, 0xD7A7A3E4, 0x7902C93D, 0x34A4A861, 0xE3030B85, 0xAEA56AD9,
	0xF204DF7A, 0xBFA2BE26, 0x68051DC2, 0x25A37C9E, 0x8B061647, 0xC6A0771B, 0x1107D4FF, 0x5CA1B5A3,
	0xA908F3F4, 0xE4AE92A8, 0x3309314C, 0x7EAF5010, 0xD00A3AC9, 0x9DAC5B95, 0x4A0BF871, 0x07AD992D,
	0x5B0C2C8E, 0x16AA4DD2, 0xC10DEE36, 0x8CAB8F6A, 0x220EE5B3, 0x6FA884EF, 0xB80F270B, 0xF5A94657,
	0x1F10ABA5, 0x52B6CAF9, 0x8511691D, 0xC8B70841, 0x66126298, 0x2BB403C4, 0xFC13A020, 0xB1B5C17C,
	0xED1474DF, 0xA0B21583, 0x7715B667, 0x3AB3D73B, 0x9416BDE2, 0xD9B0DCBE, 0x0E177F5A, 0x43B11E06,
	0xB6185851, 0xFBBE390D, 0x2C199AE9, 0x61BFFBB5, 0xCF1A916C, 0x82BCF030, 0x551B53D4, 0x18BD3288,
	0x441C872B, 0x09BAE677, 0xDE1D4593, 0x93BB24CF, 0x3D1E4E16, 0x70B82F4A, 0xA71F8CAE, 0xEAB9EDF2,
	0x3E201B07, 0x73867A5B, 0xA421D9BF, 0xE987B8E3, 0x4722D23A, 0x0A84B366, 0xDD231082, 0x908571DE,
	0xCC24C47D, 0x8182A521, 0x562506C5, 0x1B836799, 0xB5260D40, 0xF8806C1C, 0x2F27CFF8, 0x6281AEA4,
	0x9728E8F3, 0xDA8E89AF, 0x0D292A4B, 0x408F4B17, 0xEE2A21CE, 0xA38C4092, 0x742BE376, 0x398D822A,
	0x652C3789, 0x288A56D5, 0xFF2DF531, 0xB28B946D, 0x1C2EFEB4, 0x51889FE8, 0x862F3C0C, 0xCB895D50,
	0x2130B0A2, 0x6C96D1FE, 0xBB31721A, 0xF6971346, 0x5832799F, 0x159418C3, 0xC233BB27, 0x8F95DA7B,
	0xD3346FD8, 0x9E920E84, 0x4935AD60, 0x0493CC3C, 0xAA36A6E5, 0xE790C7B9, 0x3037645D, 0x7D910501,
	0x88384356, 0xC59E220A, 0x123981EE, 0x5F9FE0B2, 0xF13A8A6B, 0xBC9CEB37, 0x6B3B48D3, 0x269D298F,
	0x7A3C9C2C, 0x379AFD70, 0xE03D5E94, 0xAD9B3FC8, 0x033E5511, 0x4E98344D, 0x993F97A9, 0xD499F6F5,
	0x7C40360E, 0x31E65752, 0xE641F4B6, 0xABE795EA, 0x0542FF33, 0x48E49E6F, 0x9F433D8B, 0xD2E55CD7,
	0x8E44E974, 0xC3E28828, 0x14452BCC, 0x59E34A90, 0xF7462049, 0xBAE04115, 0x6D47E2F1, 0x20E183AD,
	0xD548C5FA, 0x98EEA4A6, 0x4F490742, 0x02EF661E, 0xAC4A0CC7, 0xE1EC6D9B, 0x364BCE7F, 0x7BEDAF23,
	0x274C1A80, 0x6AEA7BDC, 0xBD4DD838, 0xF0EBB964, 0x5E4ED3BD, 0x13E8B2E1, 0xC44F1105, 0x89E97059,
	0x63509DAB, 0x2EF6FCF7, 0xF9515F13, 0xB4F73E4F, 0x1A525496, 0x57F435CA, 0x8053962E, 0xCDF5F772,
	0x915442D1, 0xDCF2238D, 0x0B558069, 0x46F3E135, 0xE8568BEC, 0xA5F0EAB0, 0x72574954, 0x3FF12808,
	0xCA586E5F, 0x87FE0F03, 0x5059ACE7, 0x1DFFCDBB, 0xB35AA762, 0xFEFCC63E, 0x295B65DA, 0x64FD0486,
	0x385CB125, 0x75FAD079, 0xA25D739D, 0xEFFB12C1, 0x415E7818, 0x0CF81944, 0xDB5FBAA0, 0x96F9DBFC,
	0x42602D09, 0x0FC64C55, 0xD861EFB1, 0x95C78EED, 0x3B62E434, 0x76C48568, 0xA163268C, 0xECC547D0,
	0xB064F273, 0xFDC2932F, 0x2A6530CB, 0x67C35197, 0xC9663B4E, 0x84C05A12, 0x5367F9F6, 0x1EC198AA,
	0xEB68DEFD, 0xA6CEBFA1, 0x71691C45, 0x3CCF7D19, 0x926A17C0, 0xDFCC769C, 0x086BD578, 0x45CDB424,
	0x196C0187, 0x54CA60DB, 0x836DC33F, 0xCECBA263, 0x606EC8BA, 0x2DC8A9E6, 0xFA6F0A02, 0xB7C96B5E,
	0x5D7086AC, 0x10D6E7F0, 0xC7714414, 0x8AD72548, 0x24724F91, 0x69D42ECD, 0xBE738D29, 0xF3D5EC75,
	0xAF7459D6, 0xE2D2388A, 0x35759B6E, 0x78D3FA32, 0xD67690EB, 0x9BD0F1B7, 0x4C775253, 0x01D1330F,
	0xF4787558, 0xB9DE1404, 0x6E79B7E0, 0x23DFD6BC, 0x8D7ABC65, 0xC0DCDD39, 0x177B7EDD, 0x5ADD1F81,
	0x067CAA22, 0x4BDACB7E, 0x9C7D689A, 0xD1DB09C6, 0x7F7E631F, 0x32D80243, 0xE57FA1A7, 0xA8D9C0FB,
};

/**
 * AES S-Box.
 */
static const uint8_t sbox[256] =
{
	0x63, 0x7C, 0x77, 0x7B, 0xF2, 0x6B, 0x6F, 0xC5, 0x30, 0x01, 0x67, 0x2B, 0xFE, 0xD7, 0xAB, 0x76,
	0xCA, 0x82, 0xC9, 0x7D, 0xFA, 0x59, 0x47, 0xF0, 0xAD, 0xD4, 0xA2, 0xAF, 0x9C, 0xA4, 0x72, 0xC0,
	0xB7, 0xFD, 0x93, 0x26, 0x36, 0x3F, 0xF7, 0xCC, 0x34, 0xA5, 0xE5, 0xF1, 0x71, 0xD8, 0x31, 0x15,
	0x04, 0xC7, 0x23, 0xC3, 0x18, 0x96, 0x05, 0x9A, 0x07, 0x12, 0x80, 0xE2, 0xEB, 0x27, 0xB2, 0x75,
	0x09, 0x83, 0x2C, 0x1A, 0x1B, 0x6E, 0x5A, 0xA0, 0x52, 0x3B, 0xD6, 0xB3, 0x29, 0xE3, 0x2F, 0x84,
	0x53, 0xD1, 0x00, 0xED, 0x20, 0xFC, 0xB1, 0x5B, 0x6A, 0xCB, 0xBE, 0x39, 0x4A, 0x4C, 0x58, 0xCF,
	0xD0, 0xEF, 0xAA, 0xFB, 0x43, 0x4D, 0x33, 0x85, 0x45, 0xF9, 0x02, 0x7F, 0x50, 0x3C, 0x9F, 0xA8,
	0x51, 0xA3, 0x40, 0x8F, 0x92, 0x9D, 0x38, 0xF5, 0xBC, 0xB6, 0xDA, 0x21, 0x10, 0xFF, 0xF3, 0xD2,
	0xCD, 0x0C, 0x13, 0xEC, 0x5F, 0x97, 0x44, 0x17, 0xC4, 0xA7, 0x7E, 0x3D, 0x64, 0x5D, 0x19, 0x73,
	0x60, 0x81, 0x4F, 0xDC, 0x22, 0x2A, 0x90, 0x88, 0x46, 0xEE, 0xB8, 0x14, 0xDE, 0x5E, 0x0B, 0xDB,
	0xE0, 0x32, 0x3A, 0x0A, 0x49, 0x06, 0x24, 0x5C, 0xC2, 0xD3, 0xAC, 0x62, 0x91, 0x95, 0xE4, 0x79,
	0xE7, 0xC8, 0x37, 0x6D, 0x8D, 0xD5, 0x4E, 0xA9, 0x6C, 0x56, 0xF4, 0xEA, 0x65, 0x7A, 0xAE, 0x08,
	0xBA, 0x78, 0x25, 0x2E, 0x1C, 0xA6, 0xB4, 0xC6, 0xE8, 0xDD, 0x74, 0x1F, 0x4B, 0xBD, 0x8B, 0x8A,
	0x70, 0x3E, 0xB5, 0x66, 0x48, 0x03, 0xF6, 0x0E, 0x61, 0x35, 0x57, 0xB9, 0x86, 0xC1, 0x1D, 0x9E,
	0xE1, 0xF8, 0x98, 0x11, 0x69, 0xD9, 0x8E, 0x94, 0x9B, 0x1E, 0x87, 0xE9, 0xCE, 0x55, 0x28, 0xDF,
	0x8C, 0xA1, 0x89, 0x0D, 0xBF, 0xE6, 0x42, 0x68, 0x41, 0x99, 0x2D, 0x0F, 0xB0, 0x54, 0xBB, 0x16,
};

/**
 * Lookup table used in the multiplication by x in the AES field.
 */
static const uint8_t mul2[256] = {
	0x00, 0x02, 0x04, 0x06, 0x08, 0x0A, 0x0C, 0x0E, 0x10, 0x12, 0x14, 0x16, 0x18, 0x1A, 0x1C, 0x1E,
	0x20, 0x22, 0x24, 0x26, 0x28, 0x2A, 0x2C, 0x2E, 0x30, 0x32, 0x34, 0x36, 0x38, 0x3A, 0x3C, 0x3E,
	0x40, 0x42, 0x44, 0x46, 0x48, 0x4A, 0x4C, 0x4E, 0x50, 0x52, 0x54, 0x56, 0x58, 0x5A, 0x5C, 0x5E,
	0x60, 0x62, 0x64, 0x66, 0x68, 0x6A, 0x6C, 0x6E, 0x70, 0x72, 0x74, 0x76, 0x78, 0x7A, 0x7C, 0x7E,
	0x80, 0x82, 0x84, 0x86, 0x88, 0x8A, 0x8C, 0x8E, 0x90, 0x92, 0x94, 0x96, 0x98, 0x9A, 0x9C, 0x9E,
	0xA0, 0xA2, 0xA4, 0xA6, 0xA8, 0xAA, 0xAC, 0xAE, 0xB0, 0xB2, 0xB4, 0xB6, 0xB8, 0xBA, 0xBC, 0xBE,
	0xC0, 0xC2, 0xC4, 0xC6, 0xC8, 0xCA, 0xCC, 0xCE, 0xD0, 0xD2, 0xD4, 0xD6, 0xD8, 0xDA, 0xDC, 0xDE,
	0xE0, 0xE2, 0xE4, 0xE6, 0xE8, 0xEA, 0xEC, 0xEE, 0xF0, 0xF2, 0xF4, 0xF6, 0xF8, 0xFA, 0xFC, 0xFE,
	0x1B, 0x19, 0x1F, 0x1D, 0x13, 0x11, 0x17, 0x15, 0x0B, 0x09, 0x0F, 0x0D, 0x03, 0x01, 0x07, 0x05,
	0x3B, 0x39, 0x3F, 0x3D, 0x33, 0x31, 0x37, 0x35, 0x2B, 0x29, 0x2F, 0x2D, 0x23, 0x21, 0x27, 0x25,
	0x5B, 0x59, 0x5F, 0x5D, 0x53, 0x51, 0x57, 0x55, 0x4B, 0x49, 0x4F, 0x4D, 0x43, 0x41, 0x47, 0x45,
	0x7B, 0x79, 0x7F, 0x7D, 0x73, 0x71, 0x77, 0x75, 0x6B, 0x69, 0x6F, 0x6D, 0x63, 0x61, 0x67, 0x65,
	0x9B, 0x99, 0x9F, 0x9D, 0x93, 0x91, 0x97, 0x95, 0x8B, 0x89, 0x8F, 0x8D, 0x83, 0x81, 0x87, 0x85,
	0xBB, 0xB9, 0xBF, 0xBD, 0xB3, 0xB1, 0xB7, 0xB5, 0xAB, 0xA9, 0xAF, 0xAD, 0xA3, 0xA1, 0xA7, 0xA5,
	0xDB, 0xD9, 0xDF, 0xDD, 0xD3, 0xD1, 0xD7, 0xD5, 0xCB, 0xC9, 0xCF, 0xCD, 0xC3, 0xC1, 0xC7, 0xC5,
	0xFB, 0xF9, 0xFF, 0xFD, 0xF3, 0xF1, 0xF7, 0xF5, 0xEB, 0xE9, 0xEF, 0xED, 0xE3, 0xE1, 0xE7, 0xE5,
};

/**
 * Lookup table used in the multiplication by (x + 1) in the AES field.
 */
static const uint8_t mul3[256] = {
	0x00, 0x03, 0x06, 0x05, 0x0C, 0x0F, 0x0A, 0x09, 0x18, 0x1B, 0x1E, 0x1D, 0x14, 0x17, 0x12, 0x11,
	0x30, 0x33, 0x36, 0x35, 0x3C, 0x3F, 0x3A, 0x39, 0x28, 0x2B, 0x2E, 0x2D, 0x24, 0x27, 0x22, 0x21,
	0x60, 0x63, 0x66, 0x65, 0x6C, 0x6F, 0x6A, 0x69, 0x78, 0x7B, 0x7E, 0x7D, 0x74, 0x77, 0x72, 0x71,
	0x50, 0x53, 0x56, 0x55, 0x5C, 0x5F, 0x5A, 0x59, 0x48, 0x4B, 0x4E, 0x4D, 0x44, 0x47, 0x42, 0x41,
	0xC0, 0xC3, 0xC6, 0xC5, 0xCC, 0xCF, 0xCA, 0xC9, 0xD8, 0xDB, 0xDE, 0xDD, 0xD4, 0xD7, 0xD2, 0xD1,
	0xF0, 0xF3, 0xF6, 0xF5, 0xFC, 0xFF, 0xFA, 0xF9, 0xE8, 0xEB, 0xEE, 0xED, 0xE4, 0xE7, 0xE2, 0xE1,
	0xA0, 0xA3, 0xA6, 0xA5, 0xAC, 0xAF, 0xAA, 0xA9, 0xB8, 0xBB, 0xBE, 0xBD, 0xB4, 0xB7, 0xB2, 0xB1,
	0x90, 0x93, 0x96, 0x95, 0x9C, 0x9F, 0x9A, 0x99, 0x88, 0x8B, 0x8E, 0x8D, 0x84, 0x87, 0x82, 0x81,
	0x9B, 0x98, 0x9D, 0x9E, 0x97, 0x94, 0x91, 0x92, 0x83, 0x80, 0x85, 0x86, 0x8F, 0x8C, 0x89, 0x8A,
	0xAB, 0xA8, 0xAD, 0xAE, 0xA7, 0xA4, 0xA1, 0xA2, 0xB3, 0xB0, 0xB5, 0xB6, 0xBF, 0xBC, 0xB9, 0xBA,
	0xFB, 0xF8, 0xFD, 0xFE, 0xF7, 0xF4, 0xF1, 0xF2, 0xE3, 0xE0, 0xE5, 0xE6, 0xEF, 0xEC, 0xE9, 0xEA,
	0xCB, 0xC8, 0xCD, 0xCE, 0xC7, 0xC4, 0xC1, 0xC2, 0xD3, 0xD0, 0xD5, 0xD6, 0xDF, 0xDC, 0xD9, 0xDA,
	0x5B, 0x58, 0x5D, 0x5E, 0x57, 0x54, 0x51, 0x52, 0x43, 0x40, 0x45, 0x46, 0x4F, 0x4C, 0x49, 0x4A,
	0x6B, 0x68, 0x6D, 0x6E, 0x67, 0x64, 0x61, 0x62, 0x73, 0x70, 0x75, 0x76, 0x7F, 0x7C, 0x79, 0x7A,
	0x3B, 0x38, 0x3D, 0x3E, 0x37, 0x34, 0x31, 0x32, 0x23, 0x20, 0x25, 0x26, 0x2F, 0x2C, 0x29, 0x2A,
	0x0B, 0x08, 0x0D, 0x0E, 0x07, 0x04, 0x01, 0x02, 0x13, 0x10, 0x15, 0x16, 0x1F, 0x1C, 0x19, 0x1A,
};

/**
 * Constants used by MASHA.
 */
#define CONST1 0xE10C33E5
#define CONST2 0x4C7A8DCB
#define CONST3 0xC0EAD75E
#define CONST4 0x394E6808

/**
 * MASHA Sub function.
 *
 * Applies AES SubBytes followed by MixColumns.
 */
static uint32_t sub(uint32_t input) {
	uint8_t a, b, c, d;
	uint32_t t;

	a = sbox[(input) & 0xFF];
	b = sbox[(input >> 8) & 0xFF];
	c = sbox[(input >> 16) & 0xFF];
	d = sbox[(input >> 24) & 0xFF];

	t = mul2[a] ^ mul3[b] ^ c ^ d;
	t |= (uint32_t) (a ^ mul2[b] ^ mul3[c] ^ d) << 8;
	t |= (uint32_t) (a ^ b ^ mul2[c] ^ mul3[d]) << 16;
	t |= (uint32_t) (mul3[a] ^ b ^ c ^ mul2[d]) << 24;

	return t;
}

/**
 * MASHA MSub function.
 *
 * Applies AES MixColumns followed by Sub (i.e. applies MixColumns,
 * SubBytes, and MixColumns again).
 */
static uint32_t msub(uint32_t input) {
	uint8_t x, y, z, w;
	uint32_t t;

	x = (input) & 0xFF;
	y = (input >> 8) & 0xFF;
	z = (input >> 16) & 0xFF;
	w = (input >> 24) & 0xFF;

	t = mul2[x] ^ mul3[y] ^ z ^ w;
	t |= (uint32_t) (x ^ mul2[y] ^ mul3[z] ^ w) << 8;
	t |= (uint32_t) (x ^ y ^ mul2[z] ^ mul3[w]) << 16;
	t |= (uint32_t) (mul3[x] ^ y ^ z ^ mul2[w]) << 24;

	return sub(t);
}

/**
 * Encrypts or decrypts a block.

 * @param ctx			- the context.
 * @param output		- the output.
 * @param input			- the input.
 * @param input_len		- the length in bytes of the input.
 * @param encrypt		- 1 for encryption, 0 for decryption.
 */
static void process_block(ac_masha_ctx_t ctx, uint32_t *output, uint32_t *input,
		size_t input_len, int encrypt) {
	uint32_t *a;
	uint32_t *b;
	uint32_t t1, t2, t3;

	a = ctx->A + 3 + (ctx->clock & 0x7);
	b = ctx->B + 16 + (ctx->clock & 0xf);

	if (input_len == 8) {
		output[0] = a[0] ^ (ctx->L2 + ctx->L1) ^ input[0];
		output[1] = a[4] ^ (ctx->R2 + ctx->R1) ^ input[1];

		t1 = mul_alpha0[a[0] >> 24] ^ (a[0] << 8) ^ a[1] ^ a[3];
		*(a - 3) = *(a + 5) = t1;

		t1 = mul_alpha1[b[0] >> 24] ^ (b[0] << 8);
		if ((a[2] >> 16) & (1u << 15)) {
			t1 += b[1];
		} else {
			t1 ^= b[1];
		}
		t1 ^= b[4];
		if ((a[2] >> 16) & (1u << 14)) {
			t1 += b[6];
		} else {
			t1 ^= b[6];
		}
		t1 ^= b[10];
		t1 ^= (encrypt ? output[0] : input[0]);
		t1 = msub(t1) + a[3];

		*(b - 5) = *(b + 11) = t1;

		t1 = b[5] ^ (encrypt ? output[1] : input[1]);
		t1 = msub(t1) + a[1];

		t2 = ctx->R2 + b[0];
		t3 = ctx->L2 + b[5];
		ctx->R2 = sub(ctx->R1);
		ctx->L2 = sub(ctx->L1);
		ctx->R1 = sub(t3);
		ctx->L1 = sub(t2);

		*(b - 11) = *(b + 5) = t1;

		ctx->clock++;

	} else if (input_len < 8) {
		uint32_t keystream[2];
		ctx->complete = 0;
		keystream[0] = a[0] ^ (ctx->L2 + ctx->L1);
		keystream[1] = a[4] ^ (ctx->R2 + ctx->R1);
		if (encrypt) {
			bc_t plaintext;
			memcpy(plaintext, input, input_len);
			plaintext[input_len] = 0x80;
			memset(plaintext + input_len + 1, 0,
					AC_MASHA_BLOCK_LEN - input_len - 1);
			ctx->ciphertext[0] = keystream[0] ^ ((uint32_t *) plaintext)[0];
			ctx->ciphertext[1] = keystream[1] ^ ((uint32_t *) plaintext)[1];
			memcpy(output, ctx->ciphertext, input_len);
		} else {
			memcpy(ctx->ciphertext, input, input_len);
			memcpy(((unsigned char *) ctx->ciphertext) + input_len,
					((unsigned char *) keystream) + input_len,
					AC_MASHA_BLOCK_LEN - input_len);
			((unsigned char *) ctx->ciphertext)[input_len] ^= 0x80;
			keystream[0] ^= input[0];
			keystream[1] ^= input[1];
			memcpy(output, keystream, input_len);
		}
	}
}

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

void ac_masha_key(ac_masha_ctx_t ctx, unsigned char *key, size_t key_len) {
}

void ac_masha_init(ac_masha_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len) {
	uint32_t *a;
	uint32_t *b;
	uint32_t t1, t2;

	uint32_t *pkey = (uint32_t *) key;
	uint32_t *piv = (uint32_t *) iv;

	ctx->A[3] = pkey[0];
	ctx->A[4] = pkey[1];
	ctx->A[5] = 0;
	ctx->A[6] = pkey[2];
	ctx->A[7] = pkey[3];
	ctx->B[16] = piv[0];
	ctx->B[17] = piv[1];
	ctx->B[18] = piv[2];
	ctx->B[19] = 0;
	ctx->B[20] = 0;
	ctx->B[21] = piv[3];
	ctx->B[22] = piv[4];
	ctx->B[23] = piv[5];
	ctx->B[24] = 0;
	ctx->B[25] = 0;
	ctx->B[26] = 0;

	ctx->R1 = 0;
	ctx->R2 = 0;
	ctx->L1 = 0;
	ctx->L2 = 0;

	ctx->complete = 1;

	for (ctx->clock = 0; ctx->clock < 28; ctx->clock++) {
		a = ctx->A + 3 + (ctx->clock & 0x7);
		b = ctx->B + 16 + (ctx->clock & 0xF);

		t1 = mul_alpha0[a[0] >> 24] ^ (a[0] << 8) ^ a[1] ^ a[3];
		t1 ^= a[4] + (ctx->R2 + ctx->R1);

		*(a - 3) = *(a + 5) = t1;

		t2 = mul_alpha1[b[0] >> 24] ^ (b[0] << 8);
		if ((a[2] >> 16) & (1u << 15)) {
			t2 += b[1];
		} else {
			t2 ^= b[1];
		}
		t2 ^= b[4];
		if ((a[2] >> 16) & (1u << 14)) {
			t2 += b[6];
		} else {
			t2 ^= b[6];
		}
		t2 += b[10];
		t2 ^= a[0] ^ (ctx->L2 + ctx->L1);

		*(b - 5) = *(b + 11) = t2;

		t1 = ctx->R2 + b[0];
		t2 = ctx->L2 + b[5];
		ctx->R2 = sub(ctx->R1);
		ctx->L2 = sub(ctx->L1);
		ctx->R1 = sub(t2);
		ctx->L1 = sub(t1);
	}
}

void ac_masha_data(ac_masha_ctx_t ctx, unsigned char *data, size_t data_len) {
	THROW(ERR_NO_VALID);
}

void ac_masha_enc(ac_masha_ctx_t ctx, unsigned char *output,
		unsigned char *input, size_t input_len) {
	process_block(ctx, (uint32_t *) output, (uint32_t *) input, input_len, 1);
}

void ac_masha_tag(ac_masha_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	uint32_t *a;
	uint32_t *b;
	uint32_t *ptag = (uint32_t *) tag;
	uint32_t t1, t2, t3;
	uint32_t in[2];
	uint32_t out[2];
	int i, j;

	if (tag_len != AC_MASHA_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}

	in[0] = (ctx->complete ? 0x80000000 : 0);
	in[1] = 0;
	process_block(ctx, out, in, sizeof(in), 1);
	ctx->L1 = ctx->L1 ^ CONST1;
	ctx->L2 = ctx->L2 + CONST2;
	ctx->R1 = ctx->R1 ^ CONST3;
	ctx->R2 = ctx->R2 + CONST4;

	in[0] = in[1] = 0;

	for (i = 0; i < 22; i++) {
		process_block(ctx, out, in, sizeof(in), 1);
	}

	a = ctx->A + 3 + (ctx->clock & 0x7);
	b = ctx->B + 16 + (ctx->clock & 0xf);

	for (i = j = 0; i < 3; i += 1) {
		ptag[j++] = a[i] ^ (ctx->L2 + ctx->L1);
		ptag[j++] = a[4 + i] ^ (ctx->R2 + ctx->R1);
		t1 = mul_alpha0[a[i] >> 24] ^ (a[i] << 8) ^ a[1 + i] ^ a[3 + i];
		t2 = ctx->R2 + b[i];
		t3 = ctx->L2 + b[5 + i];
		ctx->R2 = sub(ctx->R1);
		ctx->L2 = sub(ctx->L1);
		ctx->R1 = sub(t3);
		ctx->L1 = sub(t2);
		a[5 + i] = t1;
	}
	ptag[6] = a[3] ^ (ctx->L2 + ctx->L1);
	ptag[7] = a[7] ^ (ctx->R2 + ctx->R1);
}

void ac_masha_dec(ac_masha_ctx_t ctx, unsigned char *msg,
		unsigned char *cipher_msg, size_t cipher_msg_len) {
	process_block(ctx, (uint32_t *) msg, (uint32_t *) cipher_msg,
			cipher_msg_len, 0);
}

int ac_masha_check(ac_masha_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	unsigned char computed_tag[AC_MASHA_TAG_LEN];

	if (tag_len != AC_MASHA_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return 0;
	}

	ac_masha_tag(ctx, computed_tag, sizeof(computed_tag));

	if (util_cmp_const(tag, computed_tag, AC_MASHA_TAG_LEN) == CMP_EQ) {
		return 1;
	} else {
		return 0;
	}
}
