/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * This code is somewhat based on code by Ted Krovetz, under the following
 * permission:
 */

/*------------------------------------------------------------------------
/ OCB Version 3 Reference Code (Optimized C)     Last modified 08-SEP-2012
/-------------------------------------------------------------------------
/ Copyright (c) 2012 Ted Krovetz.
/
/ Permission to use, copy, modify, and/or distribute this software for any
/ purpose with or without fee is hereby granted, provided that the above
/ copyright notice and this permission notice appear in all copies.
/
/ THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
/ WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
/ MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
/ ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
/ WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
/ ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
/ OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
/
/ Phillip Rogaway holds patents relevant to OCB. See the following for
/ his patent grant: http://www.cs.ucdavis.edu/~rogaway/ocb/grant.htm
/
/ Special thanks to Keegan McAllister for suggesting several good improvements
/
/ Comments are welcome: Ted Krovetz <ted@krovetz.net> - Dedicated to Laurel K
/------------------------------------------------------------------------- */

/**
 * @file
 *
 * Implementation of the OCB3 authenticated cipher functions.
 *
 * @version $Id$
 * @ingroup ac
 */

#include <stdio.h>
#include <string.h>

#include "relic_core.h"
#include "relic_types.h"
#include "relic_ac.h"
#include "relic_ac_low.h"
#include "relic_bc_low.h"
#include "relic_test.h"
#include "relic_util.h"

/*============================================================================*/
/* Private definitions                                                        */
/*============================================================================*/

#define L_TABLE_SZ 8

static int lsh(dig_t *c, dig_t *a, int size, int bits) {
	int i;
	dig_t r, carry, shift, mask;

	shift = DIGIT - bits;
	carry = 0;
	mask = MASK(bits);
	a += size - 1;
	c += size - 1;
	for (i = size - 1; i >= 0; i--, a--, c--) {
		/* Get the needed least significant bits. */
		r = ((*a) >> shift) & mask;
		/* Shift left the operand. */
		*c = ((*a) << bits) | carry;
		/* Update the carry. */
		carry = r;
	}
	return carry;
}

static int lshm(dig_t *c, dig_t *a, int size, int bits) {
	int digits;
	dig_t carry;

	SPLIT(bits, digits, bits, DIGIT_LOG);

	if (digits > 0) {
		memmove(c, a + digits, (size - digits) * sizeof(dig_t));
		memset(c + size - digits, 0, digits * sizeof(dig_t));
	} else {
		memcpy(c, a, BC_BLOCK_LEN);
	}

	if (bits > 0) {
		carry = lsh(c, c, size, bits);
	}
	return carry;
}

static void swap_if_le(bc_t c, bc_t a) {
#ifdef BIGED
	memcpy(c, a, AC_OCB3_BLOCK_LEN);
#else
	bc_t t;
	int i, j;
	for (i = 0; i < AC_OCB3_BLOCK_LEN; i += sizeof(dig_t)) {
		for (j = 0; j < sizeof(dig_t); j++) {
			t[i + j] = a[i + (sizeof(dig_t) - j - 1)];
		}
	}
	memcpy(c, t, AC_OCB3_BLOCK_LEN);
#endif
}

static unsigned char *getL(ac_ocb3_ctx_t ctx, unsigned tz) {
	unsigned char *rval;
	if (tz < L_TABLE_SZ)
		return ctx->L[tz];
	else {
		unsigned i;
		rval = ctx->Ltmp;
		/* Bring L[MAX] into registers, make it register correct */
		swap_if_le(rval, ctx->L[L_TABLE_SZ - 1]);
		ac_ocb3_dbl_low(rval);
		for (i = L_TABLE_SZ; i < tz; i++)
			ac_ocb3_dbl_low(rval);
		swap_if_le(rval, rval); /* To memory correct */
		return rval;
	}
}

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

void ac_ocb3_key(ac_ocb3_ctx_t ctx, unsigned char *key, size_t key_len) {
	bc_t tmp_blk;

#ifdef BC_SINGL
	ctx->config = 0;
#endif

	if (key_len != BC_KEY_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}

#ifdef BC_SINGL
	memcpy(ctx->key, key, key_len);
#endif
	bc_enc_key(ctx->enc_ctx, key, key_len);
#ifndef BC_SINGL
	bc_dec_key(ctx->dec_ctx, key, key_len);
#endif

	memset(ctx->Lstar, 0, BC_BLOCK_LEN);
	bc_enc_start(ctx->enc_ctx, ctx->Lstar, ctx->Lstar);
	memset(ctx->ad_offset, 0, BC_BLOCK_LEN);
	memset(ctx->cached_Top, 0, BC_BLOCK_LEN);
	bc_enc_end(ctx->enc_ctx, ctx->Lstar);

	swap_if_le(tmp_blk, ctx->Lstar);
	ac_ocb3_dbl_low(tmp_blk);
	swap_if_le(ctx->Ldollar, tmp_blk);
	ac_ocb3_dbl_low(tmp_blk);
	swap_if_le(ctx->L[0], tmp_blk);
	for (int i = 1; i < L_TABLE_SZ; i++) {
		ac_ocb3_dbl_low(tmp_blk);
		swap_if_le(ctx->L[i], tmp_blk);
	}
}

void ac_ocb3_init(ac_ocb3_ctx_t ctx, unsigned char *key, size_t key_len,
		unsigned char *iv, size_t iv_len, size_t msg_len, size_t data_len) {
	unsigned int idx;
	unsigned char *ktop = (unsigned char *) ctx->KtopStr;
	bc_t tmp_blk;

	ctx->blocks_processed = 0;
	ctx->ad_blocks_processed = 0;

	if (key_len != BC_KEY_LEN || iv_len != AC_OCB3_IV_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}

	memcpy(tmp_blk + 4, iv, AC_OCB3_IV_LEN);
	tmp_blk[3] = 1;
	tmp_blk[0] = tmp_blk[1] = tmp_blk[2] = 0;
	idx = tmp_blk[15] & 0x3f; /* Get low 6 bits of nonce  */
	tmp_blk[15] &= 0xc0; /* Zero low 6 bits of nonce */

	if (memcmp(tmp_blk, ctx->cached_Top, BC_BLOCK_LEN) != 0) {
		bc_enc_start(ctx->enc_ctx, ktop, tmp_blk);
		memcpy(ctx->cached_Top, tmp_blk, BC_BLOCK_LEN);
		bc_enc_end(ctx->enc_ctx, ktop);

		memcpy(ktop + BC_BLOCK_LEN, ktop, BC_BLOCK_LEN / 2);
		bc_nxor_low(ktop + BC_BLOCK_LEN, ktop + BC_BLOCK_LEN, ktop + 1,
				BC_BLOCK_LEN / 2 - 1);
		ktop[(3 * BC_BLOCK_LEN) / 2 - 1] ^= ktop[BC_BLOCK_LEN / 2];

		/* Initialize offset and checksum */
		swap_if_le(ktop, ktop);
		swap_if_le(ktop + BC_BLOCK_LEN, ktop + BC_BLOCK_LEN);
	}
	memset(ctx->ad_checksum, 0, BC_BLOCK_LEN);
	memset(ctx->checksum, 0, BC_BLOCK_LEN);
	lshm(ctx->KtopStr, ctx->KtopStr, ((3 * BC_BLOCK_LEN) / 2) / sizeof(dig_t),
			idx);
	swap_if_le(ctx->offset, ktop);
}

void ac_ocb3_data(ac_ocb3_ctx_t ctx, unsigned char *data, size_t data_len) {
	unsigned char *p;
	unsigned int tz;
	bc_t t;

#ifdef BC_SINGL
	if (ctx->config != 0) {
		bc_enc_key(ctx->enc_ctx, ctx->key, BC_KEY_LEN);
		ctx->config = 0;
	}
#endif

	if (data_len < BC_BLOCK_LEN) {
		bc_xor_low(ctx->ad_offset, ctx->ad_offset, ctx->Lstar);
		memset(t, 0, BC_BLOCK_LEN);
		memcpy(t, data, data_len);
		t[data_len] = 0x80;
		bc_xor_low(t, t, ctx->ad_offset);
		bc_enc(ctx->enc_ctx, t, t);
		bc_xor_low(ctx->ad_checksum, ctx->ad_checksum, t);
		return;
	}

	ctx->ad_blocks_processed += 1;
	tz = ac_ocb3_ntz_low(ctx->ad_blocks_processed);
	p = getL(ctx, tz);
	bc_xor_low(ctx->ad_offset, ctx->ad_offset, p);
	bc_xor_low(t, data, ctx->ad_offset);
	bc_enc(ctx->enc_ctx, t, t);
	bc_xor_low(ctx->ad_checksum, ctx->ad_checksum, t);
}

void ac_ocb3_enc(ac_ocb3_ctx_t ctx, unsigned char *output, unsigned char *input,
		size_t input_len) {
	unsigned char *p;
	unsigned int tz;
	bc_t t;

#ifdef BC_SINGL
	if (ctx->config != 0) {
		bc_enc_key(ctx->enc_ctx, ctx->key, BC_KEY_LEN);
		ctx->config = 0;
	}
#endif

	if (input_len < BC_BLOCK_LEN) {
		bc_xor_low(ctx->offset, ctx->offset, ctx->Lstar);
		bc_enc_start(ctx->enc_ctx, t, ctx->offset);
		bc_nxor_low(ctx->checksum, ctx->checksum, input, input_len);
		ctx->checksum[input_len] ^= 0x80;
		bc_enc_end(ctx->enc_ctx, t);
		bc_nxor_low(output, input, t, input_len);
		return;
	}

	ctx->blocks_processed += 1;
	tz = ac_ocb3_ntz_low(ctx->blocks_processed);
	p = getL(ctx, tz);
	bc_xor_low(ctx->offset, ctx->offset, p);
	bc_xor_low(t, input, ctx->offset);
	bc_enc_start(ctx->enc_ctx, t, t);
	bc_xor_low(ctx->checksum, ctx->checksum, input);
	bc_enc_end(ctx->enc_ctx, t);
	bc_xor_low(output, t, ctx->offset);
}

void ac_ocb3_dec(ac_ocb3_ctx_t ctx, unsigned char *output, unsigned char *input,
		size_t input_len) {
	unsigned char *p;
	unsigned int tz;
	bc_t t;

	if (input_len < BC_BLOCK_LEN) {
#ifdef BC_SINGL
		if (ctx->config != 0) {
			bc_enc_key(ctx->enc_ctx, ctx->key, BC_KEY_LEN);
			ctx->config = 0;
		}
#endif
		bc_xor_low(ctx->offset, ctx->offset, ctx->Lstar);
		bc_enc(ctx->enc_ctx, t, ctx->offset);
		bc_nxor_low(output, input, t, input_len);
		bc_nxor_low(ctx->checksum, ctx->checksum, output, input_len);
		ctx->checksum[input_len] ^= 0x80;
		return;
	}

#ifdef BC_SINGL
	if (ctx->config != 1) {
		bc_dec_key(ctx->dec_ctx, ctx->key, BC_KEY_LEN);
		ctx->config = 1;
	}
#endif

	ctx->blocks_processed += 1;
	tz = ac_ocb3_ntz_low(ctx->blocks_processed);
	p = getL(ctx, tz);
	bc_xor_low(ctx->offset, ctx->offset, p);
	bc_xor_low(t, input, ctx->offset);
	bc_dec(ctx->dec_ctx, t, t);
	bc_xor_low(output, t, ctx->offset);
	bc_xor_low(ctx->checksum, ctx->checksum, output);
}

void ac_ocb3_tag(ac_ocb3_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	bc_t t;

	if (tag_len != AC_OCB3_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return;
	}

#ifdef BC_SINGL
	if (ctx->config != 0) {
		bc_enc_key(ctx->enc_ctx, ctx->key, BC_KEY_LEN);
		ctx->config = 0;
	}
	memset(ctx->key, 0, BC_KEY_LEN);
#endif

	bc_xor_low(ctx->offset, ctx->offset, ctx->Ldollar);
	bc_xor_low(t, ctx->checksum, ctx->offset);
	bc_enc(ctx->enc_ctx, t, t);
	bc_xor_low(t, t, ctx->ad_checksum);
	memcpy(tag, t, AC_OCB3_TAG_LEN);
}

int ac_ocb3_check(ac_ocb3_ctx_t ctx, unsigned char *tag, size_t tag_len) {
	unsigned char computed_tag[AC_OCB3_TAG_LEN];

	if (tag_len != AC_OCB3_TAG_LEN) {
		THROW(ERR_NO_VALID);
		return 0;
	}

	ac_ocb3_tag(ctx, computed_tag, sizeof(computed_tag));

	if (util_cmp_const(tag, computed_tag, AC_OCB3_TAG_LEN) == CMP_EQ) {
		return 1;
	} else {
		return 0;
	}
}
