/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of the low-level authenticated cipher Hummingbird-2 functions.
 *
 * @version $Id$
 * @ingroup ac
 */

#include <string.h>

#include "relic_ac.h"
#include "relic_ac_low.h"
#include "relic_types.h"
#include "relic_util.h"

/*============================================================================*/
/* Private definitions                                                        */
/*============================================================================*/

/*
 * The four Hummingbird-2 S-Boxes, and their inverses.
 */
static const unsigned char S1[] = {7, 12, 14, 9, 2, 1, 5, 15, 11, 6, 13, 0, 4, 8, 10, 3};
static const unsigned char S1inv[] = {11, 5, 4, 15, 12, 6, 9, 0, 13, 3, 14, 8, 1, 10, 2, 7};
static const unsigned char S2[] = {4, 10, 1, 6, 8, 15, 7, 12, 3, 0, 14, 13, 5, 9, 11, 2};
static const unsigned char S2inv[] = {9, 2, 15, 8, 0, 12, 3, 6, 4, 13, 1, 14, 7, 11, 10, 5};
static const unsigned char S3[] = {2, 15, 12, 1, 5, 6, 10, 13, 14, 8, 3, 4, 0, 11, 9, 7};
static const unsigned char S3inv[] = {12, 3, 0, 10, 11, 4, 5, 15, 9, 14, 6, 13, 2, 7, 8, 1};
static const unsigned char S4[] = {15, 4, 5, 8, 9, 7, 2, 1, 10, 3, 0, 14, 6, 12, 13, 11};
static const unsigned char S4inv[] = {10, 7, 6, 9, 1, 2, 12, 5, 3, 4, 8, 15, 13, 14, 11, 0};

/*
 * Combined inverse S-Boxes.
 */
static const unsigned char S12inv[] = {202u, 199u, 198u, 201u, 193u, 194u, 204u, 197u, 195u, 196u, 200u, 207u, 205u, 206u, 203u, 192u, 58u, 55u, 54u, 57u, 49u, 50u, 60u, 53u, 51u, 52u, 56u, 63u, 61u, 62u, 59u, 48u, 10u, 7u, 6u, 9u, 1u, 2u, 12u, 5u, 3u, 4u, 8u, 15u, 13u, 14u, 11u, 0u, 170u, 167u, 166u, 169u, 161u, 162u, 172u, 165u, 163u, 164u,
		168u, 175u, 173u, 174u, 171u, 160u, 186u, 183u, 182u, 185u, 177u, 178u, 188u, 181u, 179u, 180u, 184u, 191u, 189u, 190u, 187u, 176u, 74u, 71u, 70u, 73u, 65u, 66u, 76u, 69u, 67u, 68u, 72u, 79u, 77u, 78u, 75u, 64u, 90u, 87u, 86u, 89u, 81u, 82u, 92u, 85u, 83u, 84u, 88u, 95u, 93u, 94u, 91u, 80u, 250u, 247u, 246u, 249u, 241u, 242u, 252u, 245u, 243u, 244u, 248u, 255u, 253u, 254u, 251u, 240u, 154u, 151u, 150u, 153u,
		145u, 146u, 156u, 149u, 147u, 148u, 152u, 159u, 157u, 158u, 155u, 144u, 234u, 231u, 230u, 233u, 225u, 226u, 236u, 229u, 227u, 228u, 232u, 239u, 237u, 238u, 235u, 224u, 106u, 103u, 102u, 105u, 97u, 98u, 108u, 101u, 99u, 100u, 104u, 111u, 109u, 110u, 107u, 96u, 218u, 215u, 214u, 217u, 209u, 210u, 220u, 213u, 211u, 212u, 216u, 223u, 221u, 222u, 219u, 208u, 42u, 39u, 38u, 41u, 33u, 34u, 44u, 37u, 35u, 36u, 40u, 47u, 45u, 46u, 43u, 32u, 122u, 119u, 118u, 121u, 113u, 114u, 124u, 117u, 115u, 116u, 120u, 127u, 125u, 126u, 123u, 112u, 138u, 135u, 134u, 137u, 129u, 130u, 140u, 133u, 131u, 132u, 136u, 143u, 141u, 142u, 139u, 128u, 26u, 23u, 22u, 25u, 17u, 18u, 28u, 21u, 19u, 20u, 24u, 31u, 29u, 30u, 27u, 16u};
static const unsigned char S34inv[] = {185u, 178u, 191u, 184u, 176u, 188u, 179u, 182u, 180u, 189u, 177u, 190u, 183u, 187u, 186u, 181u, 89u, 82u, 95u, 88u, 80u, 92u, 83u, 86u, 84u, 93u, 81u, 94u, 87u, 91u, 90u, 85u, 73u, 66u, 79u, 72u, 64u, 76u, 67u, 70u, 68u, 77u, 65u, 78u, 71u, 75u, 74u, 69u, 249u, 242u, 255u, 248u, 240u, 252u, 243u, 246u, 244u, 253u, 241u, 254u, 247u, 251u, 250u, 245u, 201u, 194u, 207u, 200u, 192u, 204u, 195u, 198u, 196u, 205u, 193u, 206u, 199u, 203u, 202u, 197u, 105u, 98u, 111u, 104u, 96u, 108u, 99u, 102u, 100u, 109u, 97u, 110u, 103u, 107u, 106u, 101u, 153u, 146u, 159u, 152u, 144u, 156u, 147u, 150u, 148u, 157u, 145u, 158u, 151u, 155u, 154u, 149u, 9u, 2u, 15u, 8u, 0u, 12u, 3u, 6u, 4u, 13u, 1u, 14u, 7u, 11u, 10u, 5u, 217u, 210u,
		223u, 216u, 208u, 220u, 211u, 214u, 212u, 221u, 209u, 222u, 215u, 219u, 218u, 213u, 57u, 50u, 63u, 56u, 48u, 60u, 51u, 54u, 52u, 61u, 49u, 62u, 55u, 59u, 58u, 53u, 233u, 226u, 239u, 232u, 224u, 236u, 227u,
		230u, 228u, 237u, 225u, 238u, 231u, 235u, 234u, 229u, 137u, 130u, 143u, 136u, 128u, 140u, 131u, 134u, 132u, 141u, 129u, 142u, 135u, 139u, 138u, 133u, 25u, 18u, 31u, 24u, 16u, 28u, 19u, 22u, 20u, 29u, 17u, 30u, 23u, 27u, 26u, 21u, 169u, 162u, 175u, 168u, 160u, 172u, 163u, 166u, 164u, 173u, 161u, 174u, 167u, 171u, 170u, 165u, 41u, 34u, 47u, 40u, 32u, 44u, 35u, 38u, 36u, 45u, 33u, 46u, 39u, 43u, 42u, 37u, 121u,
		114u, 127u, 120u, 112u, 124u, 115u, 118u, 116u, 125u, 113u, 126u, 119u, 123u, 122u, 117u};

/*
 * Precomputed inverse L function.
 */
static const uint16_t L0inv[] = {0u, 20501u, 41002u, 61503u, 16469u, 4160u, 57471u, 45162u, 32938u, 53439u, 8320u, 28821u, 49407u, 37098u, 24789u, 12480u, 341u, 20800u, 41343u, 61802u, 16640u, 4373u, 57642u, 45375u, 33279u, 53738u, 8661u, 29120u, 49578u, 37311u, 24960u, 12693u, 682u, 21183u, 41600u, 62101u, 17151u, 4842u, 58069u, 45760u, 33280u, 53781u, 8746u, 29247u, 49749u, 37440u, 25215u, 12906u, 1023u, 21482u, 41941u, 62400u, 17322u,
		5055u, 58240u, 45973u, 33621u, 54080u, 9087u, 29546u, 49920u, 37653u, 25386u, 13119u, 1364u, 21825u, 42366u, 62827u, 17665u, 5396u, 58667u, 46398u, 34302u, 54763u, 9684u, 30145u, 50603u, 38334u, 25985u, 13716u, 1025u, 21524u, 42027u, 62526u, 17492u, 5185u, 58494u, 46187u, 33963u, 54462u, 9345u, 29844u, 50430u, 38123u, 25812u, 13505u, 2046u, 22507u, 42964u, 63425u, 18347u, 6078u, 59265u, 46996u, 34644u, 55105u, 10110u, 30571u, 50945u, 38676u, 26411u, 14142u, 1707u, 22206u, 42625u, 63124u, 18174u, 5867u, 59092u, 46785u, 34305u, 54804u, 9771u, 30270u, 50772u, 38465u, 26238u, 13931u, 2728u, 23229u, 43650u, 64151u,
		19197u, 6888u, 60119u, 47810u, 35330u, 55831u, 10792u, 31293u, 51799u, 39490u, 27261u, 14952u, 3069u, 23528u, 43991u, 64450u, 19368u, 7101u, 60290u, 48023u, 35671u, 56130u, 11133u, 31592u, 51970u, 39703u, 27432u, 15165u, 2050u, 22551u, 43048u, 63549u, 18519u, 6210u, 59517u, 47208u, 34984u, 55485u, 10370u, 30871u, 51453u, 39144u, 26839u, 14530u, 2391u, 22850u, 43389u, 63848u, 18690u, 6423u, 59688u, 47421u, 35325u, 55784u, 10711u, 31170u, 51624u, 39357u, 27010u, 14743u, 4092u, 24553u, 45014u, 65475u, 20393u, 8124u, 61315u, 49046u, 36694u, 57155u, 12156u, 32617u, 52995u, 40726u, 28457u, 16188u, 3753u, 24252u, 44675u, 65174u, 20220u, 7913u, 61142u, 48835u, 36355u, 56854u, 11817u, 32316u, 52822u, 40515u, 28284u, 15977u, 3414u, 23875u, 44412u, 64873u, 19715u, 7446u, 60713u, 48444u, 36348u, 56809u, 11734u, 32195u, 52649u, 40380u, 28035u, 15766u, 3075u, 23574u, 44073u, 64572u, 19542u, 7235u, 60540u, 48233u, 36009u, 56508u, 11395u, 31894u, 52476u, 40169u, 27862u, 15555u};
static const uint16_t L1inv[] = {0u, 5456u, 10912u, 16368u, 21824u, 16400u, 32736u, 27312u, 43648u, 49104u, 32800u, 38256u, 65472u, 60048u, 54624u, 49200u, 21761u, 16465u, 32673u, 27377u, 65u, 5393u, 10977u, 16305u, 65409u, 60113u, 54561u, 49265u, 43713u, 49041u, 32865u, 38193u, 43522u, 48978u, 32930u, 38386u, 65346u, 59922u, 54754u, 49330u, 130u, 5586u, 10786u, 16242u, 21954u, 16530u, 32610u, 27186u, 65283u, 59987u, 54691u, 49395u, 43587u, 48915u, 32995u, 38323u, 21891u, 16595u, 32547u, 27251u, 195u, 5523u, 10851u, 16179u, 21509u, 16725u, 32421u, 27637u, 325u, 5141u, 11237u, 16053u, 65157u, 60373u, 54309u, 49525u, 43973u, 48789u, 33125u, 37941u, 260u, 5204u, 11172u, 16116u, 21572u, 16660u, 32484u, 27572u, 43908u, 48852u, 33060u, 38004u, 65220u, 60308u, 54372u, 49460u, 65031u, 60247u, 54439u, 49655u, 43847u, 48663u, 33255u, 38071u, 21639u, 16855u, 32295u, 27511u, 455u, 5271u, 11111u, 15927u, 43782u, 48726u, 33190u, 38134u, 65094u, 60182u, 54502u, 49590u, 390u, 5334u, 11046u, 15990u, 21702u, 16790u, 32358u, 27446u, 43018u, 48474u, 33450u, 38906u, 64842u, 59418u, 55274u, 49850u, 650u, 6106u, 10282u, 15738u, 22474u, 17050u, 32106u, 26682u, 64779u,
		59483u, 55211u, 49915u, 43083u, 48411u, 33515u, 38843u, 22411u, 17115u, 32043u, 26747u, 715u, 6043u, 10347u, 15675u, 520u, 5976u, 10408u, 15864u, 22344u, 16920u, 32232u, 26808u, 43144u, 48600u, 33320u, 38776u, 64968u, 59544u, 55144u, 49720u, 22281u, 16985u, 32169u, 26873u, 585u, 5913u, 10473u, 15801u, 64905u, 59609u, 55081u, 49785u, 43209u, 48537u, 33385u, 38713u, 64527u, 59743u, 54959u, 50175u, 43343u, 48159u, 33775u, 38591u, 22159u, 17375u, 31791u, 27007u, 975u, 5791u, 10607u, 15423u, 43278u, 48222u, 33710u, 38654u, 64590u, 59678u, 55022u, 50110u, 910u, 5854u, 10542u, 15486u, 22222u, 17310u, 31854u, 26942u, 22029u, 17245u, 31917u, 27133u, 845u, 5661u, 10733u, 15549u, 64653u, 59869u, 54829u, 50045u, 43469u, 48285u, 33645u, 38461u, 780u, 5724u, 10668u, 15612u, 22092u, 17180u, 31980u, 27068u, 43404u, 48348u, 33580u, 38524u, 64716u, 59804u, 54892u, 49980u};

/*
 * Precomputed f function.
 */
static const uint16_t f0[] = {47087u, 39204u, 40293u, 43560u, 44649u, 38375u, 32930u, 35937u, 41642u, 34019u, 34848u, 45998u, 37286u, 47916u, 49005u, 42731u, 49980u, 60919u, 59830u, 57083u, 55994u, 57652u, 62577u, 63666u, 54905u, 61488u, 64755u, 51069u, 58741u, 53247u, 52158u, 53816u, 3852u, 8647u, 9606u, 4811u, 5770u, 11524u, 14401u, 13442u, 6729u, 15360u, 12483u, 2893u, 10565u, 975u, 1934u, 7688u, 31711u, 21780u, 20821u, 26136u, 25177u, 22999u, 19602u, 16465u, 28314u, 18643u, 17424u, 32670u, 23958u, 30492u, 29533u, 27355u, 27550u, 17749u, 16660u, 30297u, 29208u, 18838u, 23763u, 20496u, 32475u, 22674u, 21585u, 28639u, 19927u, 26461u, 25372u, 31386u, 42926u, 35173u, 36132u, 47721u, 48680u, 34214u, 37091u, 39968u, 45803u, 38050u, 39009u, 41967u, 33255u, 43885u, 44844u, 46762u, 38765u, 47526u, 48615u, 35498u, 36587u, 46437u, 40992u, 44259u,
		33320u, 42081u, 43170u, 37676u, 45348u, 39854u, 40943u, 34409u, 19228u, 26071u, 24982u, 22235u, 21146u, 26900u, 31825u, 28818u, 24153u, 30736u, 29907u, 20317u, 27989u, 18399u, 17310u, 23064u, 34604u, 43495u, 44454u, 39659u, 40618u, 42276u, 45153u, 48290u, 37481u, 46112u, 47331u, 33645u, 41317u, 35823u, 36782u, 38440u, 8013u, 12678u, 13767u, 650u, 1739u, 15685u, 10240u, 9411u, 2568u, 11329u, 8322u, 6924u, 14596u, 5006u, 6095u, 3657u, 62463u, 56628u, 55669u, 60984u, 60025u, 53751u, 50354u, 51313u, 59066u, 49395u, 52272u, 63422u, 54710u, 65340u, 64381u, 58107u, 12174u, 325u, 1284u, 12873u, 13832u, 3462u, 6339u, 5120u, 15051u, 7298u, 4161u, 11215u, 2503u, 9037u, 9996u, 16010u, 16335u, 4356u, 5445u, 8712u, 9801u, 7623u, 2178u, 1089u, 10890u, 3267u, 0u, 15246u, 6534u, 13068u, 14157u, 11979u, 54141u, 64950u, 63991u, 52922u, 51963u, 61813u, 58416u, 59635u, 50744u, 57457u, 60594u, 55100u, 62772u, 57278u, 56319u, 49785u, 23389u, 30102u, 29143u, 18074u, 17115u, 31061u, 27664u, 24787u, 19992u, 26705u, 25746u, 24348u, 32020u, 22430u, 21471u, 19033u, 58302u, 52597u, 51508u, 65145u, 64056u, 49590u, 54515u, 55344u, 63227u, 53426u,
		56433u, 59391u, 50679u, 61309u, 60220u, 62138u};
static const uint16_t f1[] = {30157u, 64502u, 12504u, 63429u, 31230u, 48867u, 46785u, 32239u, 45776u, 29148u, 65511u, 15595u, 13513u, 14586u, 47858u, 62420u, 50977u, 18714u, 33332u, 17705u, 51986u, 3087u, 1069u, 52995u, 60u, 49968u,
		19723u, 36359u, 34341u, 35350u, 2078u, 16696u, 59305u, 27026u, 41660u, 26017u, 60314u, 11399u, 9381u, 61323u, 8372u, 58296u, 28035u, 44687u, 42669u, 43678u, 10390u, 25008u, 38517u, 6222u, 54112u, 5245u, 39494u, 23899u, 21881u, 40535u, 20840u, 37476u, 7263u, 57171u, 55153u, 56130u, 22858u, 4204u, 9369u, 43682u, 24972u, 42641u, 10410u, 61367u, 59285u, 11451u, 58244u, 8328u, 44723u, 28095u, 26013u, 27054u, 60326u, 41600u, 5205u, 39534u, 20800u, 38493u, 6246u, 57211u, 55129u, 7287u, 54088u, 4164u, 40575u, 23923u,
		21841u, 22882u, 56170u, 37452u, 21829u, 56190u, 4176u, 55117u, 22902u, 40555u, 38473u, 23911u, 37464u,
		20820u, 57199u, 7267u, 5185u, 6258u, 39546u, 54108u, 63469u, 31190u, 45816u, 30181u, 64478u, 15555u, 13537u, 65487u, 12528u, 62460u, 32199u, 48843u, 46825u, 47834u, 14546u, 29172u, 46845u, 14534u, 62440u, 13557u, 47822u, 32211u, 30193u, 48863u, 29152u, 45804u, 15575u, 65499u, 63481u, 64458u, 31170u, 12516u,
		25993u, 60338u, 8348u, 59265u, 27066u, 44711u, 42629u, 28075u, 41620u, 24984u, 61347u, 11439u, 9357u, 10430u, 43702u, 58256u, 55141u, 22878u, 37488u, 21869u, 56150u, 7243u, 5225u, 57159u, 4216u, 54132u, 23887u, 40515u, 38497u, 39506u, 6234u, 20860u, 1041u, 35370u, 16644u, 34329u, 2082u, 53055u, 50973u, 3123u, 49932u, 0u, 36411u, 19767u, 17685u, 18726u, 52014u, 33288u, 17665u, 52026u, 20u, 50953u, 18738u, 36399u, 34317u, 19747u, 33308u, 16656u, 53035u, 3111u, 1029u, 2102u, 35390u, 49944u, 34353u, 2058u, 49956u, 1081u, 35330u, 19743u, 17725u, 36371u, 16684u, 33312u, 3099u, 53015u, 50997u, 51974u, 18702u, 40u, 42681u, 10370u, 58284u, 9393u, 43658u, 28055u, 26037u, 44699u, 24996u, 41640u, 11411u, 61343u, 59325u, 60302u, 27014u, 8352u, 13533u, 47846u, 29128u, 46805u, 14574u, 65523u, 63441u, 15615u, 62400u, 12492u, 48887u, 32251u, 30169u, 31210u, 64482u, 45764u};

/*
 * There are many choices for precomputation in Hummingbird-2.
 * Some code is commented out below; they provide alternative ways to computing
 * some functions.
 */

#if 0
static uint16_t S(uint16_t x) {
	return S4[x & 0xF]
	          | (S3[(x >> 4) & 0xF] << 4)
	          | (S2[(x >> 8) & 0xF] << 8)
	          | (S1[(x >> 12) & 0xF] << 12);
}
#endif

static uint16_t Sinv(uint16_t x) {
	return S12inv[x & 0xFF] | (S34inv[(x >> 8) & 0xFF] << 8);
}

#if 0
static uint16_t L(uint16_t x) {
	return x ^ (x << 6) ^ (x >> 10) ^ (x << 10) ^ (x >> 6);
}
#endif

static uint16_t Linv(uint16_t x) {
	unsigned char x0 = (unsigned char) x;
	unsigned char x1 = (unsigned char) (x >> 8);
	return L0inv[x0] ^ L1inv[x1];
//	return x ^ (x << 4) ^ (x >> 12) ^ (x << 2) ^ (x >> 14)
//			 ^ (x >> 4) ^ (x << 12) ^ (x >> 2) ^ (x << 14);
}

static uint16_t f(uint16_t x) {
	unsigned char x0 = (unsigned char) x;
	unsigned char x1 = (unsigned char) (x >> 8);
	return f0[x0] ^ f1[x1];
//	return L(S(x));
}

static uint16_t finv(uint16_t x) {
	return Sinv(Linv(x));
}

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

uint16_t ac_hb2_wd16_low(uint16_t x, uint16_t b, uint16_t c, uint16_t d) {
	return f(f(f(f(x) ^ b) ^ c) ^ d);
}

uint16_t ac_hb2_wd16inv_low(uint16_t x, uint16_t b, uint16_t c, uint16_t d) {
	return finv(finv(finv(finv(x) ^ d) ^ c) ^ b);
}
