/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Implementation of the low-level authenticated cipher GCM functions.
 *
 * @version $Id$
 * @ingroup ac
 */

#include "relic_bc_conf.h"

#ifndef AC_GCM_SGCM

NAME relic_fb_mul_low_asm
RSEG CODE
EVEN

PUBLIC ac_gcm_muln_low
PUBLIC ac_gcm_rdcn_low
PUBLIC ac_gcm_tab_low
EXTERN ac_gcm_lshbtab_low


/*============================================================================*/
/* Private definitions                                                        */
/*============================================================================*/

//#define KARATSUBA

/*
 * Pointer register which points to the parameter a.
 */
#define A_PTR       r13

/*
 * Pointer register which points to the parameter b.
 */
#define B_PTR       r14

/*
 * Pointer register which points to the parameter c.
 */
#define C_PTR       r12

/*
 * Temporary register.
 */
#define RT          r4

/*
 * Pointer to the table.
 */
#define T_PTR       r15

/*
 * Size in bits of the multiplication of a binary field element by a polynomial
 * with degree lower than 4.
 */
#define POLY_M4     (FB_POLYN + 4)

/*
 * Size in digits of a precomputation table line.
 */
#define T_LINE 9

/*
 * Size of the multiplication precomputation table in digits.
 */
#define T_DIGS  (16 * T_LINE)

PROLOGUE MACRO
    pushm.w #8,r11
 ENDM

MULN_TABLE MACRO
    FILL_TABLE 0, 7 /* fill 8 columns */
    PREP_LAST   r8, r9, r10, r11 /* fill last column */
    FILL_COLUMN r8, r9, r10, r11
 ENDM

FILL_TABLE MACRO i, j
    IF i = 0
        PREP_FIRST  r8, r9, r10, r11
        FILL_COLUMN r8, r9, r10, r11
    ELSE
        PREP_COLUMN i + i, r8, r9, r10, r11
        FILL_COLUMN r8, r9, r10, r11
    ENDIF
    incd T_PTR
    IF i < j
        FILL_TABLE i + 1, j
    ENDIF
 ENDM

/*
 * Prepares the first column values.
 */
PREP_FIRST MACRO r1, r2, r4, r8
    mov     @B_PTR+, r1    ; r1 = b[0]
    mov     r1, r2
    rla     r2             ; r2 = r1 << 1
    mov     r2, r4
    rla     r4             ; r4 = r1 << 2
    mov     r4, r8
    rla     r8             ; r8 = r1 << 3
 ENDM

/*
 * Prepares the values for column i.
 */
PREP_COLUMN MACRO i, r1, r2, r4, r8
    mov     r1, RT

    mov     @B_PTR+, r1
    mov     r1, r2
    rla     RT
    rlc     r2
    mov     r2, r4
    rla     RT
    rlc     r4
    mov     r4, r8
    rla     RT
    rlc     r8
 ENDM

/*
 * Prepares the values for the last column.
 */
PREP_LAST MACRO r1, r2, r4, r8
    mov     r1, RT

    clr     r1
    mov     r1, r2
    rla     RT
    rlc     r2
    mov     r2, r4
    rla     RT
    rlc     r4
    mov     r4, r8
    rla     RT
    rlc     r8
 ENDM

/*
 * Fills a column of the precomputation table.
 */
FILL_COLUMN MACRO r1, r2, r4, r8
    clr     0(T_PTR)
    mov     r1, 1*2*T_LINE(T_PTR)  ; tab[1][i] = r1
    mov     r2, 2*2*T_LINE(T_PTR)  ; tab[2][i] = r2
    mov     r1, RT
    xor     r2, RT
    mov     RT, 3*2*T_LINE(T_PTR)   ; tab[3][i] = r1^r2
    mov     r4, 4*2*T_LINE(T_PTR)  ; tab[4][i] = r4
    xor     r4, RT
    mov     RT, 7*2*T_LINE(T_PTR)   ; tab[7][i] = r1^r2^r4
    xor     r1, RT
    mov     RT, 6*2*T_LINE(T_PTR)   ; tab[6][i] = r2^r4
    mov     r1, RT
    xor     r4, RT
    mov     RT, 5*2*T_LINE(T_PTR)   ; tab[5][i] = r1^r4
    mov     r8, 8*2*T_LINE(T_PTR)  ; tab[8][i] = r8
    mov     r1, RT
    xor     r8, RT
    mov     RT, 9*2*T_LINE(T_PTR)   ; tab[9][i] = r1^r8
    xor     r2, RT
    mov     RT, 11*2*T_LINE(T_PTR)  ; tab[11][i] = r1^r2^r8
    xor     r1, RT
    mov     RT, 10*2*T_LINE(T_PTR)  ; tab[10][i] = r2^r8
    xor     r8, r4
    mov     r4, 12*2*T_LINE(T_PTR); tab[12][i] = r4^r8
    xor     r1, r4
    mov     r4, 13*2*T_LINE(T_PTR); tab[13][i] = r1^r4^r8
    xor     r2, r4
    mov     r4, 15*2*T_LINE(T_PTR); tab[15][i] = r1^r2^r4^r8
    xor     r1, r4
    mov     r4, 14*2*T_LINE(T_PTR); tab[14][i] = r2^r4^r8
 ENDM

EPILOGUE MACRO
    popm.w #8,r11
    reta
 ENDM

/*============================================================================*/
/* Public definitions                                                         */
/*============================================================================*/

ac_gcm_tab_low:
    /* r12: table, r13: b */
    PROLOGUE

    mov r13,r14
    mov r12, T_PTR

    MULN_TABLE

    calla #ac_gcm_lshbtab_low

    EPILOGUE

ac_gcm_muln_low:
    /* r12: c
       r13: a
       r14: table
    */
    /* pushes */
    PROLOGUE

    sub #2,r1

#include "fb_mul_128_ld.inc"

    add #2,r1

    /* pops */
    EPILOGUE


#ifdef AC_GCM_REFLC

REDC_COL MACRO i
    //reduce into a[i]
    //r10 pending a[i+1] write

    mov (2 * (7 - i))(r13),r11

    mov r11,r14

    clr r15
    rrum.w #1,r14
    rrc r15
    xor r14,r11
    xor r15,r10

    rrum.w #1,r14
    rrc r15
    xor r14,r11
    xor r15,r10

    rrum.w #1,r14
    rrc r15
    rrum.w #1,r14
    rrc r15
    rrum.w #1,r14
    rrc r15
    rrum.w #1,r14
    rrc r15
    rrum.w #1,r14
    rrc r15
    xor r14,r11
    xor r15,r10

    IF i < 7
        //write a[i+1]
        mov r10,(2 * (7 - i - 1))(r12)
    ELSE
        xor r10,(2 * (15 - i - 1))(r13)
    ENDIF

    mov r11,r10

    //r10: a[i]
    xor (2 * (15 - i))(r13),r10
 ENDM

#else

REDC_COL MACRO i
    //reduce into a[i]
    //r10 pending a[i+1] write

    mov (2 * (8 + i))(r13),r11

    mov r11,r14

    clr r15
    rla r14
    rlc r15
    xor r14,r11
    xor r15,r10

    rla r14
    rlc r15
    xor r14,r11
    xor r15,r10

    rla r14
    rlc r15
    rla r14
    rlc r15
    rla r14
    rlc r15
    rla r14
    rlc r15
    rla r14
    rlc r15
    xor r14,r11
    xor r15,r10

    IF i < 7
        //write a[i+1]
        mov r10,(2 * (i + 1))(r12)
    ELSE
        xor r10,(2 * (i + 1))(r13)
    ENDIF

    mov r11,r10

    //r10: a[i]
    xor (2 * i)(r13),r10
 ENDM

#endif

ac_gcm_rdcn_low:
    /* r12: c
       r13: a
    */
    push.w r10
    push.w r11

#ifdef AC_GCM_REFLC
    rla 2*0(r13)
    rlc 2*1(r13)
    rlc 2*2(r13)
    rlc 2*3(r13)
    rlc 2*4(r13)
    rlc 2*5(r13)
    rlc 2*6(r13)
    rlc 2*7(r13)
    rlc 2*8(r13)
    rlc 2*9(r13)
    rlc 2*10(r13)
    rlc 2*11(r13)
    rlc 2*12(r13)
    rlc 2*13(r13)
    rlc 2*14(r13)
    rlc 2*15(r13)
#endif

    clr r10

    REDC_COL 7
    REDC_COL 6
    REDC_COL 5
    REDC_COL 4
    REDC_COL 3
    REDC_COL 2
    REDC_COL 1
    REDC_COL 0

#ifdef AC_GCM_REFLC
    mov r10,2*7(r12)
#else
    mov r10,0(r12)
#endif

    pop.w r11
    pop.w r10
    reta

#endif /* #ifndef AC_GCM_SGCM */

END
