#include "relic_bc_conf.h"

#if BC_AE == GCM

/*============================================================================*/
/* Private definitions                                                        */
/*============================================================================*/

//#define KARATSUBA

/*
 * Pointer register which points to the parameter a.
 */
#define A_PTR       r14

/*
 * Pointer register which points to the parameter b.
 */
#define B_PTR       r13

/*
 * Pointer register which points to the parameter c.
 */
#define C_PTR       r15

/*
 * Temporary register.
 */
#define RT          r4

/*
 * Pointer to the table.
 */
#define T_PTR       r12

/*
 * Size in bits of the multiplication of a binary field element by a polynomial
 * with degree lower than 4.
 */
#define POLY_M4     (FB_POLYN + 4)

/*
 * Size in digits of a precomputation table line.
 */
#define T_LINE 9

/*
 * Size of the multiplication precomputation table in digits.
 */
#define T_DIGS  (16 * T_LINE)

.macro PROLOGUE
    push r4
    push r5
    push r6
    push r7
    push r8
    push r9
    push r10
    push r11
.endm

.macro MULN_TABLE
    FILL_TABLE 0, 7 /* fill 8 columns */
    PREP_LAST   r8, r9, r10, r11 /* fill last column */
    FILL_COLUMN r8, r9, r10, r11
.endm

.macro FILL_TABLE i, j
    .if \i == 0
        PREP_FIRST  r8, r9, r10, r11
        FILL_COLUMN r8, r9, r10, r11
    .else
        PREP_COLUMN \i + \i, r8, r9, r10, r11
        FILL_COLUMN r8, r9, r10, r11
    .endif
    incd T_PTR
    .if \i < \j
        FILL_TABLE \i + 1, \j
    .endif
.endm

/*
 * Prepares the first column values.
 */
.macro PREP_FIRST r1, r2, r4, r8
    mov     @B_PTR+, \r1    ; r1 = b[0]
    mov     \r1, \r2
    rla     \r2             ; r2 = r1 << 1
    mov     \r2, \r4
    rla     \r4             ; r4 = r1 << 2
    mov     \r4, \r8
    rla     \r8             ; r8 = r1 << 3
.endm

/*
 * Prepares the values for column i.
 */
.macro PREP_COLUMN i, r1, r2, r4, r8
    mov     \r1, RT

    mov     @B_PTR+, \r1
    mov     \r1, \r2
    rla     RT
    rlc     \r2
    mov     \r2, \r4
    rla     RT
    rlc     \r4
    mov     \r4, \r8
    rla     RT
    rlc     \r8
.endm

/*
 * Prepares the values for the last column.
 */
.macro PREP_LAST r1, r2, r4, r8
    mov     \r1, RT

    clr     \r1
    mov     \r1, \r2
    rla     RT
    rlc     \r2
    mov     \r2, \r4
    rla     RT
    rlc     \r4
    mov     \r4, \r8
    rla     RT
    rlc     \r8
.endm

/*
 * Fills a column of the precomputation table.
 */
.macro FILL_COLUMN r1, r2, r4, r8
    clr     0(T_PTR)
    mov     \r1, 1*2*T_LINE(T_PTR)  ; tab[1][i] = r1
    mov     \r2, 2*2*T_LINE(T_PTR)  ; tab[2][i] = r2
    mov     \r1, RT
    xor     \r2, RT
    mov     RT, 3*2*T_LINE(T_PTR)   ; tab[3][i] = r1^r2
    mov     \r4, 4*2*T_LINE(T_PTR)  ; tab[4][i] = r4
    xor     \r4, RT
    mov     RT, 7*2*T_LINE(T_PTR)   ; tab[7][i] = r1^r2^r4
    xor     \r1, RT
    mov     RT, 6*2*T_LINE(T_PTR)   ; tab[6][i] = r2^r4
    mov     \r1, RT
    xor     \r4, RT
    mov     RT, 5*2*T_LINE(T_PTR)   ; tab[5][i] = r1^r4
    mov     \r8, 8*2*T_LINE(T_PTR)  ; tab[8][i] = r8
    mov     \r1, RT
    xor     \r8, RT
    mov     RT, 9*2*T_LINE(T_PTR)   ; tab[9][i] = r1^r8
    xor     \r2, RT
    mov     RT, 11*2*T_LINE(T_PTR)  ; tab[11][i] = r1^r2^r8
    xor     \r1, RT
    mov     RT, 10*2*T_LINE(T_PTR)  ; tab[10][i] = r2^r8
    xor     \r8, \r4
    mov     \r4, 12*2*T_LINE(T_PTR); tab[12][i] = r4^r8
    xor     \r1, \r4
    mov     \r4, 13*2*T_LINE(T_PTR); tab[13][i] = r1^r4^r8
    xor     \r2, \r4
    mov     \r4, 15*2*T_LINE(T_PTR); tab[15][i] = r1^r2^r4^r8
    xor     \r1, \r4
    mov     \r4, 14*2*T_LINE(T_PTR); tab[14][i] = r2^r4^r8
.endm

.macro EPILOGUE
    pop r11
    pop r10
    pop r9
    pop r8
    pop r7
    pop r6
    pop r5
    pop r4
    ret
.endm

.text
.align 2

.global fb_muln_low
.global fb_muln_table

fb_muln_table:
    /* r15: table, r14: b */
    PROLOGUE

    mov r14,r13
    mov r15, T_PTR

    MULN_TABLE

    EPILOGUE
    ret

fb_muln_low:
    /* r15: c
       r14: a
       r13: table
    */
    /* pushes */
    PROLOGUE

    sub #2,r1

#include "fb_mul_128_ld.inc"

    add #2,r1

    /* pops */
    EPILOGUE

#endif
