#if 1
/*
 * RELIC is an Efficient LIbrary for Cryptography
 * Copyright (C) 2007-2012 RELIC Authors
 *
 * This file is part of RELIC. RELIC is legal property of its developers,
 * whose names are not listed here. Please refer to the COPYRIGHT file
 * for contact information.
 *
 * RELIC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * RELIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with RELIC. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * Tests for block ciphers.
 *
 * @version $Id$
 * @ingroup test
 */

#ifdef __ICC430__
#include <msp430.h>
#endif

#include <stdio.h>

#include "relic.h"
#include "relic_test.h"

void dump(void *p, int len) {
	unsigned char *a = p;
	for (int i = 0; i < len; i++) {
		printf("%02X", a[i]);
	}
	puts("");
}

static int aes128(void) {
	align unsigned char keys[][BC_AES128_KEY_LEN] = {
			{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
			{0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
	};
	align unsigned char plaintexts[][BC_AES128_BLOCK_LEN] = {
			{0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
			{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
	};
	align unsigned char ciphertexts[][BC_AES128_BLOCK_LEN] = {
			{0x3A, 0xD7, 0x8E, 0x72, 0x6C, 0x1E, 0xC0, 0x2B, 0x7E, 0xBF, 0xE9, 0x2B, 0x23, 0xD9, 0xEC, 0x34},
			{0x0E, 0xDD, 0x33, 0xD3, 0xC6, 0x21, 0xE5, 0x46, 0x45, 0x5B, 0xD8, 0xBA, 0x14, 0x18, 0xBE, 0xC8},
	};
	int code = STS_ERR;
	bc_ctx_t ctx;
	bc_t t;

	TEST_ONCE("AES-128 is correct") {
		for (int k = 0; k < (sizeof(keys) / sizeof(keys[0])); k++) {
			memset(t, 0, sizeof(t));
			bc_enc_key(ctx, keys[k], sizeof(keys[k]));
			bc_enc(ctx, t, plaintexts[k]);
			TEST_ASSERT(memcmp(t, ciphertexts[k], sizeof(t)) == 0, end);

			memset(t, 0, sizeof(t));
			bc_dec_key(ctx, keys[k], sizeof(keys[k]));
			bc_dec(ctx, t, ciphertexts[k]);
			TEST_ASSERT(memcmp(t, plaintexts[k], sizeof(t)) == 0, end);
		}
	}
	TEST_END;

	code = STS_OK;

  end:
	return code;
}

static int block_cipher(void) {
	int code = STS_ERR;
	unsigned char key[BC_KEY_LEN];
	bc_t t1, t2, t3;
	bc_ctx_t ctx;

	TEST_BEGIN("block cipher is correct") {
		rand_bytes(key, sizeof(key));
		rand_bytes(t1, sizeof(t1));
		bc_enc_key(ctx, key, sizeof(key));
		bc_enc(ctx, t2, t1);
		bc_dec_key(ctx, key, sizeof(key));
		bc_dec(ctx, t3, t2);
		TEST_ASSERT(memcmp(t1, t3, sizeof(t1)) == 0, end);
	}
	TEST_END;

	code = STS_OK;

  end:
	return code;
}

int main(void) {
#ifdef __ICC430__
	// Stop watchdog timer to prevent time out reset
	WDTCTL = WDTPW + WDTHOLD;

	UCSCTL3 |= SELREF_2;// Set DCO FLL reference = REFO
	UCSCTL4 |= SELA_2;// Set ACLK = REFO

	__bis_SR_register(SCG0);// Disable the FLL control loop
	UCSCTL0 = 0x0000;// Set lowest possible DCOx, MODx
	UCSCTL1 = DCORSEL_7;// Select DCO range 24MHz operation
	UCSCTL2 = FLLD_1 + 609;// Set DCO Multiplier for 20 MHz
						   // (N + 1) * FLLRef = Fdco
						   // (609 + 1) * 32768 = 20 MHz
						   // Set FLL Div = fDCOCLK/2
	__bic_SR_register(SCG0);// Enable the FLL control loop

	// Worst-case settling time for the DCO when the DCO range bits have been
	// changed is n x 32 x 32 x f_MCLK / f_FLL_reference. See UCS chapter in 5xx
	// UG for optimization.
	// 32 x 32 x 20 MHz / 32,768 Hz = 375000 = MCLK cycles for DCO to settle
	__delay_cycles(625000);
#endif
	core_init();

	util_banner("Tests for the BC module:\n", 0);

	if (aes128() != STS_OK) {
		core_clean();
		return 1;
	}

	if (block_cipher() != STS_OK) {
		core_clean();
		return 1;
	}

	core_clean();
	return 0;
}
#endif
